from datetime import datetime
from mage_ai.api.operations.constants import META_KEY_LIMIT, META_KEY_OFFSET
from mage_ai.api.resources.DatabaseResource import DatabaseResource
from mage_ai.data_preparation.models.constants import PipelineType
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.schedules import BlockRun, PipelineRun
from mage_ai.orchestration.pipeline_scheduler import start_scheduler, stop_pipeline_run
from sqlalchemy.orm import selectinload


class PipelineRunResource(DatabaseResource):
    datetime_keys = ['execution_date']
    model_class = PipelineRun

    @classmethod
    @safe_db_query
    def collection(self, query_arg, meta, user, **kwargs):
        pipeline_schedule_id = None
        parent_model = kwargs.get('parent_model')
        if parent_model:
            pipeline_schedule_id = parent_model.id

        backfill_id = query_arg.get('backfill_id', [None])
        if backfill_id:
            backfill_id = backfill_id[0]

        pipeline_uuid = query_arg.get('pipeline_uuid', [None])
        if pipeline_uuid:
            pipeline_uuid = pipeline_uuid[0]

        status = query_arg.get('status', [None])
        if status:
            status = status[0]

        order_by_arg = query_arg.get('order_by[]', [None])
        if order_by_arg:
            order_by_arg = order_by_arg[0]

        order_by = None
        if order_by_arg:
            order_by = []
            for s in order_by_arg.split(','):
                parts = s.strip().split(' ')
                if len(parts) >= 2:
                    order_by.append((parts[0], parts[1]))
                else:
                    order_by.append((parts[0], 'asc'))

        results = (
            PipelineRun.
            query.
            options(selectinload(PipelineRun.block_runs)).
            options(selectinload(PipelineRun.pipeline_schedule))
        )

        if backfill_id is not None:
            results = results.filter(PipelineRun.backfill_id == int(backfill_id))
        if pipeline_schedule_id is not None:
            results = results.filter(PipelineRun.pipeline_schedule_id == int(pipeline_schedule_id))
        if pipeline_uuid is not None:
            results = results.filter(PipelineRun.pipeline_uuid == pipeline_uuid)
        if status is not None:
            results = results.filter(PipelineRun.status == status)

        if order_by:
            arr = []
            for tup in order_by:
                col, asc_desc = tup
                asc_desc = asc_desc.lower()
                pr_col = getattr(PipelineRun, col)
                arr.append(getattr(pr_col, asc_desc)())
            initial_results = results.order_by(*arr)
        else:
            initial_results = \
                results.order_by(PipelineRun.execution_date.desc(), PipelineRun.id.desc())

        return initial_results

    @classmethod
    @safe_db_query
    async def process_collection(self, query_arg, meta, user, **kwargs):
        total_results = self.collection(query_arg, meta, user, **kwargs)
        total_count = total_results.count()

        limit = int(meta.get(META_KEY_LIMIT, self.DEFAULT_LIMIT))
        offset = int(meta.get(META_KEY_OFFSET, 0))

        results = total_results.limit(limit + 1).offset(offset).all()

        pipeline_schedule_id = None
        parent_model = kwargs.get('parent_model')
        if parent_model:
            pipeline_schedule_id = parent_model.id

        pipeline_uuid = query_arg.get('pipeline_uuid', [None])
        if pipeline_uuid:
            pipeline_uuid = pipeline_uuid[0]

        if meta.get(META_KEY_LIMIT, None) is not None and \
            total_results.count() >= 1 and \
                (pipeline_uuid is not None or pipeline_schedule_id is not None):

            first_result = results[0]
            first_execution_date = first_result.execution_date
            additional_results = total_results.filter(
                PipelineRun.execution_date == first_execution_date,
            )
            first_additional_results = additional_results.first()
            filter_dates = []
            if first_additional_results.id != first_result.id:
                filter_dates.append(first_execution_date)

            last_result = results[-1]
            last_execution_date = last_result.execution_date
            additional_results = total_results.filter(
                PipelineRun.execution_date == last_execution_date,
            )
            filter_dates.append(last_result.execution_date)
            results = list(
                filter(
                    lambda x: x.execution_date not in filter_dates,
                    results,
                ),
            ) + additional_results.all()

        results_size = len(results)
        has_next = results_size > limit
        final_end_idx = results_size - 1 if has_next else results_size

        result_set = self.build_result_set(
            results[0:final_end_idx],
            user,
            **kwargs,
        )

        result_set.metadata = {
            'count': total_count,
            'next': has_next,
        }

        return result_set

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        pipeline_schedule = kwargs.get('parent_model')

        pipeline = Pipeline.get(pipeline_schedule.pipeline_uuid)

        if 'variables' not in payload:
            payload['variables'] = {}

        payload['pipeline_schedule_id'] = pipeline_schedule.id
        payload['pipeline_uuid'] = pipeline_schedule.pipeline_uuid
        if payload.get('execution_date') is None:
            payload['execution_date'] = datetime.utcnow()

        is_integration = PipelineType.INTEGRATION == pipeline.type
        if is_integration:
            payload['create_block_runs'] = False

        def _create_callback(resource):
            pipeline_run = resource.model
            start_scheduler(pipeline_run)

        self.on_create_callback = _create_callback

        return super().create(payload, user, **kwargs)

    @safe_db_query
    def update(self, payload, **kwargs):
        if 'retry_blocks' == payload.get('pipeline_run_action') and \
                PipelineRun.PipelineRunStatus.COMPLETED != self.model.status:
            self.model.refresh()

            pipeline = Pipeline.get(self.model.pipeline_uuid)

            incomplete_block_runs = \
                list(
                    filter(
                        lambda br: br.status != BlockRun.BlockRunStatus.COMPLETED,
                        self.model.block_runs
                    )
                )

            # Update block run status to INITIAL
            BlockRun.batch_update_status(
                [b.id for b in incomplete_block_runs],
                BlockRun.BlockRunStatus.INITIAL,
            )

            from mage_ai.orchestration.execution_process_manager \
                import execution_process_manager

            if PipelineType.STREAMING != pipeline.type:
                if PipelineType.INTEGRATION == pipeline.type:
                    execution_process_manager.terminate_pipeline_process(self.model.id)
                else:
                    for br in incomplete_block_runs:
                        execution_process_manager.terminate_block_process(
                            self.model.id,
                            br.id,
                        )

            return super().update(dict(status=PipelineRun.PipelineRunStatus.RUNNING))
        elif PipelineRun.PipelineRunStatus.CANCELLED == payload.get('status'):
            pipeline = Pipeline.get(
                self.model.pipeline_uuid,
                check_if_exists=True,
            )

            stop_pipeline_run(self.model, pipeline)

        return self



from mage_ai.shared.hash import merge_dict
from mage_ai.api.resources.DatabaseResource import DatabaseResource
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.schedules import (
    EventMatcher,
    PipelineSchedule,
    pipeline_schedule_event_matcher_association_table,
)
from sqlalchemy.orm import selectinload
import uuid


class PipelineScheduleResource(DatabaseResource):
    datetime_keys = ['start_time']
    model_class = PipelineSchedule

    @classmethod
    @safe_db_query
    def collection(self, query_arg, meta, user, **kwargs):
        pipeline = kwargs.get('parent_model')

        if pipeline:
            return (
                PipelineSchedule.
                query.
                options(selectinload(PipelineSchedule.event_matchers)).
                options(selectinload(PipelineSchedule.pipeline_runs)).
                filter(PipelineSchedule.pipeline_uuid == pipeline.uuid).
                order_by(PipelineSchedule.id.desc(), PipelineSchedule.start_time.desc())
            )

        query = PipelineSchedule.query
        order_by = query_arg.get('order_by', [None])
        if order_by[0]:
            order_by = order_by[0]
            if order_by == 'created_at':
                query = query.order_by(PipelineSchedule.created_at.desc())
            elif order_by == 'name':
                query = query.order_by(PipelineSchedule.name.asc())
            elif order_by == 'pipeline_uuid':
                query = query.order_by(PipelineSchedule.pipeline_uuid.asc())
            elif order_by == 'status':
                query = query.order_by(PipelineSchedule.status.asc())
            elif order_by == 'schedule_type':
                query = query.order_by(PipelineSchedule.schedule_type.asc())

        return query.all()

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        pipeline = kwargs['parent_model']
        payload['pipeline_uuid'] = pipeline.uuid

        if 'token' not in payload:
            payload['token'] = uuid.uuid4().hex

        return super().create(payload, user, **kwargs)

    @safe_db_query
    def update(self, payload, **kwargs):
        arr = payload.pop('event_matchers', None)
        event_matchers = []
        if arr is not None:
            if len(arr) >= 1:
                event_matchers = EventMatcher.upsert_batch(
                    [merge_dict(p, dict(pipeline_schedule_ids=[self.id])) for p in arr],
                )

            ems = (
                EventMatcher.
                query.
                join(
                    pipeline_schedule_event_matcher_association_table,
                    EventMatcher.id ==
                    pipeline_schedule_event_matcher_association_table.c.event_matcher_id
                ).
                join(
                    PipelineSchedule,
                    PipelineSchedule.id ==
                    pipeline_schedule_event_matcher_association_table.c.pipeline_schedule_id
                ).
                filter(
                    PipelineSchedule.id == int(self.id),
                    EventMatcher.id.not_in([em.id for em in event_matchers]),
                )
            )
            for em in ems:
                new_ids = [schedule for schedule in em.pipeline_schedules if schedule.id != self.id]
                ps = [p for p in PipelineSchedule.query.filter(PipelineSchedule.id.in_(new_ids))]
                em.update(pipeline_schedules=ps)

        super().update(payload)

        return self



from mage_ai.orchestration.notification.config import AlertOn, NotificationConfig
from mage_ai.services.email.email import send_email
from mage_ai.services.slack.slack import send_slack_message
from mage_ai.services.teams.teams import send_teams_message
from mage_ai.services.google_chat.google_chat import send_google_chat_message
import os


class NotificationSender:
    def __init__(self, config: NotificationConfig):
        self.config = config

    def send(
        self,
        message: str = None,
        email_subject: str = None,
        email_content: str = None,
    ) -> None:
        if message is None:
            return
        if self.config.slack_config is not None and self.config.slack_config.is_valid:
            send_slack_message(self.config.slack_config, message)

        if self.config.teams_config is not None and self.config.teams_config.is_valid:
            send_teams_message(self.config.teams_config, message)

        if self.config.google_chat_config is not None and self.config.google_chat_config.is_valid:
            send_google_chat_message(self.config.google_chat_config, message)

        if self.config.email_config is not None and email_subject is not None:
            send_email(
                self.config.email_config,
                subject=email_subject,
                message=email_content or message,
            )

    def send_pipeline_run_success_message(self, pipeline, pipeline_run) -> None:
        if AlertOn.PIPELINE_RUN_SUCCESS in self.config.alert_on:
            message = (
                f'Successfully ran Pipeline `{pipeline.uuid}` '
                f'with Trigger {pipeline_run.pipeline_schedule.id} '
                f'`{pipeline_run.pipeline_schedule.name}` '
                f'at execution time `{pipeline_run.execution_date}`.'
            )
            email_content = f'{message}\n'
            if os.getenv('ENV') != 'production':
                email_content += f'Open {self.__pipeline_run_url(pipeline, pipeline_run)} '\
                                  'to check pipeline run results and logs.'
            self.send(
                message=message,
                email_subject=f'Successfully ran Pipeline {pipeline.uuid}',
                email_content=email_content,
            )

    def send_pipeline_run_failure_message(
        self,
        pipeline,
        pipeline_run,
        message: str = None,
    ) -> None:
        if AlertOn.PIPELINE_RUN_FAILURE in self.config.alert_on:
            message = message or (
                f'Failed to run Pipeline `{pipeline.uuid}` '
                f'with Trigger {pipeline_run.pipeline_schedule.id} '
                f'`{pipeline_run.pipeline_schedule.name}` '
                f'at execution time `{pipeline_run.execution_date}`.'
            )
            email_content = f'{message}\n'
            if os.getenv('ENV') != 'production':
                email_content += f'Open {self.__pipeline_run_url(pipeline, pipeline_run)} '\
                                  'to check pipeline run results and logs.'
            self.send(
                message=message,
                email_subject=f'Failed to run Mage pipeline {pipeline.uuid}',
                email_content=email_content,
            )

    def send_pipeline_run_sla_passed_message(self, pipeline, pipeline_run) -> None:
        if AlertOn.PIPELINE_RUN_PASSED_SLA in self.config.alert_on:
            message = (
                f'SLA passed for pipeline `{pipeline.uuid}` '
                f'with Trigger {pipeline_run.pipeline_schedule.id} '
                f'`{pipeline_run.pipeline_schedule.name}` '
                f'at execution time `{pipeline_run.execution_date}`.'
            )
            email_content = f'{message}\n'
            if os.getenv('ENV') != 'production':
                email_content += f'Open {self.__pipeline_run_url(pipeline, pipeline_run)} '\
                                  'to check pipeline run results and logs.'
            self.send(
                message=message,
                email_subject=f'SLA passed for Mage pipeline {pipeline.uuid}',
                email_content=email_content,
            )

    def __pipeline_run_url(self, pipeline, pipeline_run):
        return f'http://localhost:6789/pipelines/{pipeline.uuid}/triggers/'\
               f'{pipeline_run.pipeline_schedule_id}'



from jinja2 import Template
from mage_ai.data_preparation.models.block.sql.constants import (
    CONFIG_KEY_UPSTREAM_BLOCK_CONFIGURATION,
    CONFIG_KEY_UPSTREAM_BLOCK_CONFIGURATION_TABLE_NAME,
)
from mage_ai.data_preparation.models.constants import BlockLanguage, BlockType
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.data_preparation.variable_manager import get_variable
from mage_ai.io.config import ConfigFileLoader
from os import path
from pandas import DataFrame
from typing import Callable, Dict, List, Tuple
import re


def build_variable_pattern(variable_name: str):
    return r'{}[ ]*{}[ ]*{}'.format(r'\{\{', variable_name, r'\}\}')


def blocks_in_query(block, query: str) -> Dict:
    blocks = {}

    if not query:
        return blocks

    for idx, upstream_block in enumerate(block.upstream_blocks):
        pattern = build_variable_pattern(f'df_{idx + 1}')

        if re.findall(pattern, query):
            blocks[upstream_block.uuid] = upstream_block

    return blocks


def should_cache_data_from_upstream(
    block,
    upstream_block,
    config_keys: List[str],
    config_profile_keys: List[str],
) -> bool:
    if BlockType.SENSOR == upstream_block.type:
        return False

    if BlockType.DBT == block.type:
        # TODO (tommy dang): check to see if the upstream block has the same data source
        return True

    config_path = path.join(get_repo_path(), 'io_config.yaml')

    config1 = block.configuration or {}
    config2 = upstream_block.configuration or {}

    data_provider1 = config1.get('data_provider_profile')
    data_provider2 = config2.get('data_provider_profile')

    # if config1.get('use_raw_sql'):
    #     return False

    if BlockLanguage.SQL == block.language and BlockLanguage.SQL != upstream_block.language:
        return True

    loader1 = ConfigFileLoader(config_path, data_provider1)
    loader2 = ConfigFileLoader(config_path, data_provider2)

    return not all([
        config1.get(k) and
        config2.get(k) and
        config1.get(k) == config2.get(k) for k in config_keys
    ]) or not all([
        loader1.config.get(k) and
        loader2.config.get(k) and
        loader1.config.get(k) == loader2.config.get(k) for k in config_profile_keys
    ])


def interpolate_input(
    block,
    query: str,
    replace_func: Callable = None,
    get_database: Callable = None,
    get_schema: Callable = None,
    get_table: Callable = None,
) -> str:
    def __replace_func(db, schema, tn):
        if replace_func:
            return replace_func(db, schema, tn)

        return '.'.join(list(filter(lambda x: x, [db, schema, tn])))

    for idx, upstream_block in enumerate(block.upstream_blocks):
        matcher1 = '{} df_{} {}'.format('{{', idx + 1, '}}')

        is_sql = BlockLanguage.SQL == upstream_block.language
        if is_sql:
            configuration = upstream_block.configuration
        else:
            configuration = block.configuration
        use_raw_sql = configuration.get('use_raw_sql')

        data_provider1 = block.configuration.get('data_provider')
        data_provider2 = upstream_block.configuration.get('data_provider')
        is_same_data_providers = data_provider1 == data_provider2

        if block.configuration.get('use_raw_sql'):
            if is_sql and data_provider1 != data_provider2:
                raise Exception(
                    f'Variable interpolation when using raw SQL for {matcher1} '
                    'is not supported because '
                    f'upstream block {upstream_block.uuid} is using a different '
                    f'data provider ({data_provider2}) than {block.uuid}’s data provider '
                    f'({data_provider1}). Please disable using raw SQL and try again.',
                )

        database, schema, table_name = table_name_parts(block.configuration, upstream_block)

        config_to_use = configuration if is_same_data_providers else block.configuration
        if not database:
            database = config_to_use.get('data_provider_database')
        if not schema:
            schema = config_to_use.get('data_provider_schema')

        if not database and get_database:
            database = get_database(dict(configuration=configuration))

        if not schema and get_schema:
            schema = get_schema(dict(configuration=configuration))

        if not table_name:
            table_name = upstream_block.table_name

        if get_table:
            table_name = get_table(dict(
                configuration=configuration,
                table=table_name,
            ))

        replace_with = __replace_func(database, schema, table_name)

        upstream_block_content = upstream_block.content
        if is_sql and \
                use_raw_sql and \
                is_same_data_providers and not \
                has_create_or_insert_statement(upstream_block_content):

            upstream_query = interpolate_input(upstream_block, upstream_block_content)
            replace_with = f"""(
    {upstream_query}
) AS {table_name}"""

        query = re.sub(
            build_variable_pattern(f'df_{idx + 1}'),
            replace_with,
            query,
        )

        query = query.replace(
            f'{matcher1}',
            replace_with,
        )

    return query


def interpolate_vars(query, global_vars=dict()):
    return Template(query).render(**global_vars)


def table_name_parts(
    configuration: Dict,
    upstream_block,
    no_schema: bool = False,
) -> Tuple[str, str, str]:
    database = None
    schema = None
    table = None

    full_table_name = configuration.get(
        CONFIG_KEY_UPSTREAM_BLOCK_CONFIGURATION,
        {},
    ).get(
        upstream_block.uuid,
        {},
    ).get(CONFIG_KEY_UPSTREAM_BLOCK_CONFIGURATION_TABLE_NAME)

    if full_table_name:
        parts = full_table_name.split('.')
        if len(parts) == 3:
            database, schema, table = parts
        elif len(parts) == 2:
            schema, table = parts
        elif len(parts) == 1:
            table = parts[0]

    if not table:
        table = upstream_block.table_name

    if not schema and not no_schema:
        schema = configuration.get('data_provider_schema')

    return database, schema, table


def create_upstream_block_tables(
    loader,
    block,
    cascade_on_drop: bool = False,
    configuration: Dict = None,
    execution_partition: str = None,
    cache_upstream_dbt_models: bool = False,
    cache_keys: List[str] = [],
    no_schema: bool = False,
    query: str = None,
    schema_name: str = None,
):
    from mage_ai.data_preparation.models.block.dbt.utils import (
        parse_attributes,
        source_table_name_for_block,
    )
    configuration = configuration if configuration else block.configuration

    mapping = blocks_in_query(block, query)
    for idx, upstream_block in enumerate(block.upstream_blocks):
        if query and upstream_block.uuid not in mapping:
            continue

        if should_cache_data_from_upstream(block, upstream_block, [
            'data_provider',
        ], cache_keys):
            if BlockType.DBT == upstream_block.type and not cache_upstream_dbt_models:
                continue

            df = get_variable(
                upstream_block.pipeline.uuid,
                upstream_block.uuid,
                'output_0',
                partition=execution_partition,
            )

            no_data = False
            if type(df) is DataFrame:
                if len(df.index) == 0:
                    no_data = True
            elif type(df) is dict and len(df) == 0:
                no_data = True
            elif type(df) is list and len(df) == 0:
                no_data = True
            elif not df:
                no_data = True

            if no_data:
                print(f'\n\nNo data in upstream block {upstream_block.uuid}.')
                continue

            _, schema, table_name = table_name_parts(
                configuration,
                upstream_block,
                no_schema=no_schema,
            )

            if not schema and not no_schema:
                schema = schema_name

            if BlockType.DBT == block.type and BlockType.DBT != upstream_block.type:
                if not no_schema:
                    attributes_dict = parse_attributes(block)
                    schema = attributes_dict['source_name']
                table_name = source_table_name_for_block(upstream_block)

            full_table_name = '.'.join(list(filter(lambda x: x, [
                schema,
                table_name,
            ])))

            print(f'\n\nExporting data from upstream block {upstream_block.uuid} '
                  f'to {full_table_name}.')

            loader.export(
                df,
                table_name=table_name,
                schema_name=schema,
                cascade_on_drop=cascade_on_drop,
                drop_table_on_replace=True,
                if_exists='replace',
                index=False,
                verbose=False,
            )


def extract_and_replace_text_between_strings(
    text: str,
    start_string: str,
    end_string: str = None,
    replace_string: str = '',
    case_sensitive: bool = True,
) -> Tuple[str, str]:
    start_match = re.search(start_string, text, re.NOFLAG if not case_sensitive else re.IGNORECASE)
    if end_string:
        end_match = re.search(end_string, text, re.NOFLAG if not case_sensitive else re.IGNORECASE)
    else:
        end_match = None

    if not start_match or (end_string and not end_match):
        return None, text

    start_idx = start_match.span()[0]
    if end_string and end_match:
        end_idx = end_match.span()[1]

    extracted_text = text[start_idx:end_idx]

    new_text = text[0:max(start_idx - 1, 0)] + replace_string + text[end_idx + 1:]

    return extracted_text, new_text


def remove_comments(text: str) -> str:
    lines = text.split('\n')
    return '\n'.join(line for line in lines if not line.startswith('--'))


def extract_create_statement_table_name(text: str) -> str:
    create_table_pattern = r'create table(?: if not exists)*'

    statement_partial, _ = extract_and_replace_text_between_strings(
        remove_comments(text),
        create_table_pattern,
        r'\(',
    )
    if not statement_partial:
        return None

    match1 = re.match(create_table_pattern, statement_partial, re.IGNORECASE)
    if match1:
        idx_start, idx_end = match1.span()
        new_statement = statement_partial[0:idx_start] + statement_partial[idx_end:]
        match2 = re.search(r'[^\s]+', new_statement.strip())
        if match2:
            return match2.group(0)

    parts = statement_partial[:len(statement_partial) - 1].strip().split(' ')
    return parts[-1]


def extract_insert_statement_table_names(text: str) -> List[str]:
    matches = re.findall(
        r'insert(?: overwrite)*(?: into)*[\s]+([\w.]+)',
        remove_comments(text),
        re.IGNORECASE,
    )
    return matches


def has_create_or_insert_statement(text: str) -> bool:
    table_name = extract_create_statement_table_name(text)
    if table_name:
        return True

    matches = extract_insert_statement_table_names(text)
    return len(matches) >= 1



from mage_ai.data_integrations.logger.utils import print_logs_from_output
from mage_ai.data_integrations.utils.config import (
    build_config_json,
    get_catalog,
    interpolate_variables,
)
from mage_ai.data_integrations.utils.parsers import parse_logs_and_json
from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.constants import BlockType
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.pipelines.utils import number_string
from mage_ai.data_preparation.variable_manager import get_global_variables
from mage_ai.shared.array import find
from mage_ai.shared.hash import dig
from mage_ai.shared.utils import clean_name
from mage_ai.shared.parsers import encode_complex, extract_json_objects
from typing import Any, Dict, List
import importlib
import json
import os
import pandas as pd
import simplejson
import subprocess
import yaml

PYTHON = 'python3'


class IntegrationPipeline(Pipeline):
    @property
    def blocks(self) -> List[Block]:
        return list(self.blocks_by_uuid.values())

    @property
    def data_loader(self) -> Block:
        return find(lambda x: BlockType.DATA_LOADER == x.type, self.blocks)

    @property
    def data_exporter(self) -> Block:
        return find(lambda x: BlockType.DATA_EXPORTER == x.type, self.blocks)

    @property
    def destination_config(self) -> Dict:
        if self.data_exporter and self.data_exporter.content:
            return yaml.safe_load(self.data_exporter.content)
        return {}

    @property
    def destination_uuid(self) -> str:
        return self.destination_config.get('destination')

    @property
    def destination_dir(self) -> str:
        path = f'{self.pipeline_variables_dir}/{self.destination_uuid}'
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    @property
    def destination(self) -> Any:
        if self.destination_uuid:
            return importlib.import_module(
                f'mage_integrations.destinations.{self.destination_uuid}',
            )

    @property
    def destination_file_path(self) -> str:
        try:
            if self.destination:
                return os.path.abspath(self.destination.__file__)
        except Exception:
            if self.destination_uuid:
                mod1 = importlib.import_module('mage_integrations.destinations')
                absolute_path = '/'.join(mod1.__file__.split('/')[:-1])
                absolute_path = f'{absolute_path}/{self.destination_uuid}/__init__.py'
                return absolute_path

    @property
    def source_config(self) -> Dict:
        if self.data_loader and self.data_loader.content:
            return yaml.safe_load(self.data_loader.content)
        return {}

    @property
    def source_uuid(self) -> str:
        return self.source_config.get('source')

    @property
    def source_dir(self) -> str:
        path = f'{self.pipeline_variables_dir}/{self.source_uuid}'
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    @property
    def source(self) -> Any:
        if self.source_uuid:
            return importlib.import_module(f'mage_integrations.sources.{self.source_uuid}')

    @property
    def source_file_path(self) -> str:
        try:
            if self.source:
                return os.path.abspath(self.source.__file__)
        except Exception:
            if self.source_uuid:
                mod1 = importlib.import_module('mage_integrations.sources')
                absolute_path = '/'.join(mod1.__file__.split('/')[:-1])
                absolute_path = f'{absolute_path}/{self.source_uuid}/__init__.py'
                return absolute_path

    @property
    def settings_file_path(self) -> str:
        if self.data_integration and 'catalog' in self.data_integration:
            if os.path.exists(self.catalog_config_path):
                return self.catalog_config_path
            return self.config_path
        return self.data_loader.file_path

    @property
    def transformer_file_path(self) -> str:
        transformer_file = importlib.import_module('mage_integrations.transformers.base')
        return os.path.abspath(transformer_file.__file__)

    def destination_state_file_path(self, stream: str, destination_table: str) -> str:
        stream_dir = f'{self.destination_dir}/{clean_name(stream)}'
        file_path = f'{stream_dir}/{clean_name(destination_table)}_state'

        os.makedirs(stream_dir, exist_ok=True)

        if not os.path.exists(file_path):
            with open(file_path, 'w') as f:
                f.write(json.dumps(dict(bookmarks={})))

        return file_path

    def source_output_folder(self, stream: str) -> str:
        return f'{self.source_dir}/{clean_name(stream)}/output'

    def catalog_file_path(self, stream: str, index: int = 0) -> str:
        pass

    def source_output_file_path(self, stream: str, index: int = 0) -> str:
        stream_dir = self.source_output_folder(stream)
        file_path = f'{stream_dir}/{number_string(index)}'

        os.makedirs(stream_dir, exist_ok=True)

        if not os.path.exists(file_path):
            with open(file_path, 'w') as f:
                f.write('')

        return file_path

    def source_state_file_path(self, stream: str, destination_table: str) -> str:
        stream_dir = f'{self.source_dir}/{clean_name(stream)}'
        file_path = f'{stream_dir}/{clean_name(destination_table)}_state.json'

        os.makedirs(stream_dir, exist_ok=True)

        if not os.path.exists(file_path):
            with open(file_path, 'w') as f:
                f.write(json.dumps(dict(bookmarks={})))

        return file_path

    def test_connection(self, block_type: BlockType, config: str = None):
        file_path = None
        if BlockType.DATA_LOADER == block_type:
            file_path = self.source_file_path
        elif BlockType.DATA_EXPORTER == block_type:
            file_path = self.destination_file_path

        try:
            if file_path:
                run_args = [
                    PYTHON,
                    file_path,
                    '--config_json',
                    simplejson.dumps(
                        interpolate_variables(config, self.__global_variables()),
                        default=encode_complex,
                        ignore_nan=True,
                    ),
                    '--test_connection',
                ]

                proc = subprocess.run(
                    run_args,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    timeout=10,
                )
                proc.check_returncode()
        except subprocess.CalledProcessError as e:
            stderr = e.stderr.decode('utf-8').split('\n')

            json_object = {}
            for line in stderr:
                if line.startswith('ERROR'):
                    json_object = next(extract_json_objects(line))

            error = dig(json_object, 'tags.error')
            raise Exception(error)

    def preview_data(self, block_type: BlockType, streams: List[str] = None) -> List[str]:
        from mage_integrations.utils.logger.constants import TYPE_SAMPLE_DATA

        file_path = None
        if BlockType.DATA_LOADER == block_type:
            file_path = self.source_file_path
        elif BlockType.DATA_EXPORTER == block_type:
            file_path = self.destination_file_path

        streams_updated = set()
        try:
            streams = streams if streams else \
                list(map(lambda s: s['tap_stream_id'], self.streams()))
            if file_path and len(streams) > 0:
                run_args = [
                    PYTHON,
                    file_path,
                    '--config_json',
                    build_config_json(self.data_loader.file_path, self.__global_variables()),
                    '--load_sample_data',
                    '--log_to_stdout',
                    '1',
                    '--settings',
                    self.settings_file_path,
                    '--selected_streams_json',
                    json.dumps(streams),
                ]

                proc = subprocess.run(
                    run_args,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                proc.check_returncode()

                output = proc.stdout.decode()
                print_logs_from_output(output)

                pipeline = Pipeline(self.uuid)
                block = pipeline.get_block(self.data_loader.uuid)

                for line in output.split('\n'):
                    try:
                        data = json.loads(line)
                        if TYPE_SAMPLE_DATA == data.get('type'):
                            sample_data_json = data.get('sample_data')
                            sample_data = pd.DataFrame.from_dict(json.loads(sample_data_json))
                            stream_id = data.get('stream_id')

                            variables = {
                                f'output_sample_data_{stream_id}': sample_data,
                            }

                            block.store_variables(variables)
                            streams_updated.add(stream_id)
                    except json.decoder.JSONDecodeError:
                        pass

            return streams_updated
        except subprocess.CalledProcessError as e:
            stderr = e.stderr.decode('utf-8').split('\n')

            json_object = {}
            for line in stderr:
                if line.startswith('ERROR'):
                    json_object = next(extract_json_objects(line))

            error = dig(json_object, 'tags.error')
            if not error:
                raise Exception('The sample data was not able to be loaded. Please check \
                                if the stream still exists. If it does not, click the "View and \
                                select streams" button and confirm the valid streams.')
            raise Exception(error)

    def count_records(self) -> List[Dict]:
        arr = []

        if self.source_file_path and self.data_loader.file_path:
            for stream_data in self.streams():
                tap_stream_id = stream_data['tap_stream_id']
                destination_table = stream_data.get('destination_table', tap_stream_id)

                try:
                    run_args = [
                        PYTHON,
                        self.source_file_path,
                        '--config_json',
                        build_config_json(self.data_loader.file_path, self.__global_variables()),
                        '--settings',
                        self.settings_file_path,
                        '--state',
                        self.source_state_file_path(
                            destination_table=destination_table,
                            stream=tap_stream_id,
                        ),
                        '--selected_streams_json',
                        json.dumps([tap_stream_id]),
                        '--count_records',
                    ]

                    proc = subprocess.run(run_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    proc.check_returncode()

                    arr += json.loads(parse_logs_and_json(proc.stdout.decode()))
                except subprocess.CalledProcessError as e:
                    message = e.stderr.decode('utf-8')
                    raise Exception(message)

        return arr

    def discover(self, streams: List[str] = None) -> Dict:
        if self.source_file_path and self.data_loader.file_path:
            try:
                run_args = [
                    PYTHON,
                    self.source_file_path,
                    '--config_json',
                    build_config_json(self.data_loader.file_path, self.__global_variables()),
                    '--discover',
                ]
                if streams:
                    run_args += [
                        '--selected_streams_json',
                        json.dumps(streams),
                    ]

                proc = subprocess.run(run_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                proc.check_returncode()

                return json.loads(parse_logs_and_json(proc.stdout.decode()))

            except subprocess.CalledProcessError as e:
                message = e.stderr.decode('utf-8')
                raise Exception(message)

    def discover_streams(self) -> List[str]:
        if self.source_file_path and self.data_loader.file_path:
            try:
                run_args = [
                    PYTHON,
                    self.source_file_path,
                    '--config_json',
                    build_config_json(self.data_loader.file_path, self.__global_variables()),
                    '--discover',
                    '--discover_streams',
                ]

                proc = subprocess.run(run_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                proc.check_returncode()

                return json.loads(parse_logs_and_json(proc.stdout.decode()))
            except subprocess.CalledProcessError as e:
                message = e.stderr.decode('utf-8')
                raise Exception(message)

    def streams(self, variables: Dict = {}) -> List[Dict]:
        return self.__catalog(variables)['streams']

    def __catalog(self, variables: Dict = {}) -> Dict:
        return get_catalog(
            self.data_loader,
            self.__global_variables(variables),
            pipeline=self,
        )

    def __global_variables(self, variables: Dict = {}) -> Dict:
        d = get_global_variables(self.uuid) or dict()
        d.update(variables)
        return d



import json
import yaml


def parse_logs_and_json(input_string: str) -> str:
    logs = []
    lines = []

    for line in input_string.split('\n'):
        is_log = False
        try:
            data = json.loads(line)
            is_log = type(data) is dict and 'LOG' == data.get('type')
        except json.decoder.JSONDecodeError:
            pass
        if is_log:
            logs.append(line)
        else:
            lines.append(line)

    for log in logs:
        print(log)

    return ''.join(lines)


class NoDatesSafeLoader(yaml.FullLoader):
    @classmethod
    def remove_implicit_resolver(cls, tag_to_remove):
        """
        Remove implicit resolvers for a particular tag

        Takes care not to modify resolvers in super classes.

        We want to load datetimes as strings, not dates, because we
        go on to serialise as json which doesn't have the advanced types
        of yaml, and leads to incompatibilities down the track.
        """
        if 'yaml_implicit_resolvers' not in cls.__dict__:
            cls.yaml_implicit_resolvers = cls.yaml_implicit_resolvers.copy()

        for first_letter, mappings in cls.yaml_implicit_resolvers.items():
            cls.yaml_implicit_resolvers[first_letter] = [(tag, regexp)
                                                         for tag, regexp in mappings
                                                         if tag != tag_to_remove]


NoDatesSafeLoader.remove_implicit_resolver('tag:yaml.org,2002:timestamp')



from mage_ai.data_preparation.models.block.sql import split_query_string
from mage_ai.tests.base_test import TestCase


class BlockTest(TestCase):
    def test_split_query_string(self):
        query_string = """
COPY public.temp_table FROM 's3://bucket/object_key/filename.csv'
  CREDENTIALS 'aws_access_key_id=abc123;aws_secret_access_key=efg456'
  CSV
  IGNOREHEADER 1;

INSERT INTO schema_1.table_v1
    SELECT 4 AS id
    UNION ALL
    SELECT 6 AS id;

-- CREATE TABLE
-- COMMENT


INSERT INTO schema_2.table_v2 SELECT 3 AS id UNION ALL
SELECT 5 AS id;
"""
        queries = split_query_string(query_string)

        self.assertEqual(len(queries), 3)

        self.assertEqual(
            queries[0],
            """COPY public.temp_table FROM 's3://bucket/object_key/filename.csv'
  CREDENTIALS 'aws_access_key_id=abc123;aws_secret_access_key=efg456'
  CSV
  IGNOREHEADER 1""")

        self.assertEqual(queries[1], """INSERT INTO schema_1.table_v1
    SELECT 4 AS id
    UNION ALL
    SELECT 6 AS id""")

        self.assertEqual(queries[2], """INSERT INTO schema_2.table_v2 SELECT 3 AS id UNION ALL
SELECT 5 AS id""")



from mage_ai.data_preparation.models.block import Block
from mage_ai.extensions.constants import (
    EXTENSION_UUIDS,
    EXTENSION_UUID_GREAT_EXPECTATIONS,
)
from mage_ai.data_preparation.models.block.extension.great_expectations import GreatExpectations
from typing import Dict, List


class ExtensionBlock(Block):
    def post_process_output(self, output: Dict) -> List:
        return output

    def _block_decorator(self, decorated_functions):
        def custom_code(extension_name: str, *args, **kwargs):
            if not extension_name or type(extension_name) is not str:
                raise Exception(
                    'No @extension decorator found with a valid extension name, ' +
                    "must be in the following format: @extension('extension_name')",
                )

            if extension_name not in EXTENSION_UUIDS:
                raise Exception(
                    f"Extension '{extension_name}' in @extension decorator must be 1 of: "
                    f"{', '.join(EXTENSION_UUIDS)}",
                )

            def inner(function):
                def func(*args, **kwargs):
                    if EXTENSION_UUID_GREAT_EXPECTATIONS == extension_name:
                        ge = GreatExpectations(self)
                        validators_and_uuids = ge.build_validators(*args, **kwargs)
                        validators = [t[0] for t in validators_and_uuids]
                        function(*validators)

                        validation_results = []
                        for validator, uuid in validators_and_uuids:
                            validation_result = validator.validate()
                            for result in validation_result.results:
                                if result.get('success', False):
                                    print(
                                        f'Expectations from extension {self.uuid} for ' +
                                        f'block {uuid} succeeded.',
                                    )
                                else:
                                    raise Exception(
                                        f'Expectations from extension {self.uuid} for ' +
                                        f'block {uuid} failed:\n{result}\n',
                                    )

                            validation_results.append(validation_result)

                        return validation_results

                    return function(*args, **kwargs)
                decorated_functions.append(func)

            return inner

        return custom_code



from mage_ai.data_cleaner.cleaning_rules.base import BaseRule
from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis
import numpy as np


class RemoveCollinearColumns(BaseRule):
    EPSILON = 1e-15
    MIN_ENTRIES = 3
    VIF_UB = 5

    default_config = dict(
        vif_ub=3,
    )

    def __init__(self, df, column_types, statistics, custom_config={}):
        super().__init__(df, column_types, statistics, custom_config=custom_config)
        self.numeric_df, self.numeric_columns = self._filter_numeric_types()
        self.numeric_indices = np.arange(len(self.numeric_df))

    def evaluate(self):
        suggestions = []
        if self.numeric_df.empty or len(self.numeric_df) < self.MIN_ENTRIES:
            return suggestions

        sigma = self.numeric_df.cov().to_numpy()
        std = self.numeric_df.std().to_numpy()
        pairwise_std = std * np.expand_dims(std, axis=1)
        C = sigma / (pairwise_std + self.EPSILON)

        collinear_columns = []
        good_columns = self.numeric_columns.copy()
        while True:
            e_vals = np.linalg.eigvalsh(C)
            vifs = np.sign(e_vals) / (abs(e_vals) + self.EPSILON)
            collinearity = vifs >= self.VIF_UB

            if len(collinearity) == 0:
                break
            i = collinearity.argmax()
            if i == 0 and collinearity[0] == 0:
                break
            else:
                C = np.delete(C, i, axis=0)
                C = np.delete(C, i, axis=1)
                collinear_columns.append(good_columns.pop(i))

        if len(collinear_columns) != 0:
            suggestions.append(
                self._build_transformer_action_suggestion(
                    'Remove collinear columns',
                    'Delete these columns to remove redundant data and increase data quality.',
                    ActionType.REMOVE,
                    action_arguments=collinear_columns,
                    axis=Axis.COLUMN,
                )
            )
        return suggestions



import time


def retry(
    retries: int = 3,
    delay: int = 5,
    max_delay: int = 60,
    exponential_backoff: bool = True,
):
    """
    Create the decorator to retry a method

    Args:
        retries (int, optional): Number of retry attempts.
        delay (int, optional): Delay between retries. If using exponential backoff retry,
            delay will be multiplied by 2 for each additional retry.
        max_delay (int, optional): Maximum delay time.
        exponential_backoff (bool, optional): Whether to use exponential backoff retry.
    """
    def retry_decorator(func):
        def retry_func(*args, **kwargs):
            attempt = 0
            total_delay = 0
            curr_delay = delay
            while attempt < retries:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    print(
                        'Exception thrown when attempting to run %s, attempt '
                        '%d of %d' % (func, attempt, retries)
                    )
                    attempt += 1
                    if attempt >= retries or total_delay + curr_delay >= max_delay:
                        raise e
                    time.sleep(curr_delay)
                    total_delay += curr_delay
                    if exponential_backoff:
                        curr_delay *= 2
            return func(*args, **kwargs)
        return retry_func
    return retry_decorator



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.FilePresenter import FilePresenter


class FilePolicy(BasePolicy):
    pass


FilePolicy.allow_actions([
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

FilePolicy.allow_actions([
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_edit_access())

FilePolicy.allow_read(FilePresenter.default_attributes + [], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())

FilePolicy.allow_read(FilePresenter.default_attributes + [], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

FilePolicy.allow_write([
    'api_key',
    'dir_path',
    'file',
    'name',
    'overwrite',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_edit_access())

FilePolicy.allow_write([
    'dir_path',
    'name',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_edit_access())



from mage_ai.streaming.constants import SourceType
from typing import Dict


class SourceFactory:
    @classmethod
    def get_source(self, config: Dict, **kwargs):
        connector_type = config['connector_type']
        if connector_type == SourceType.AMAZON_SQS:
            from mage_ai.streaming.sources.amazon_sqs import AmazonSqsSource

            return AmazonSqsSource(config, **kwargs)
        elif connector_type == SourceType.KAFKA:
            from mage_ai.streaming.sources.kafka import KafkaSource

            return KafkaSource(config, **kwargs)
        elif connector_type == SourceType.KINESIS:
            from mage_ai.streaming.sources.kinesis import KinesisSource

            return KinesisSource(config, **kwargs)
        elif connector_type == SourceType.AZURE_EVENT_HUB:
            from mage_ai.streaming.sources.azure_event_hub import AzureEventHubSource

            return AzureEventHubSource(config, **kwargs)
        elif connector_type == SourceType.RABBITMQ:
            from mage_ai.streaming.sources.rabbitmq import RabbitMQSource

            return RabbitMQSource(config, **kwargs)
        raise Exception(
            f'Consuming data from {connector_type} is not supported in streaming pipelines yet.',
        )



# flake8: noqa
import os

SERVER_HOST = os.getenv('HOST', 'localhost')
SERVER_PORT = os.getenv('PORT', 5789)

DATA_PREP_SERVER_PORT = os.getenv('DATA_PREP_SERVER_PORT', 6789)

DATAFRAME_OUTPUT_SAMPLE_COUNT = 10

# The last line of this file must be the version number.
# Dockerfile depends on it because it runs ./scripts/install_mage.sh and uses
# the last line to determine the version to install.
VERSION = \
'0.8.60'



from mage_ai.shared.parsers import encode_complex
from typing import Dict
import dask.dataframe as dd
import pandas as pd
import simplejson
import traceback

MAX_PARTITION_BYTE_SIZE = 100 * 1024 * 1024
JSON_SERIALIZABLE_COLUMN_TYPES = [
    dict.__name__,
    list.__name__,
]
STRING_SERIALIZABLE_COLUMN_TYPES = [
    'ObjectId',
]

CAST_TYPE_COLUMN_TYPES = set([
    'Int64',
    'int64',
    'float64',
])


def serialize_columns(row: pd.Series, column_types: Dict) -> pd.Series:
    for column, column_type in column_types.items():
        if column_type in JSON_SERIALIZABLE_COLUMN_TYPES:
            val = row[column]
            if val is not None:
                row[column] = simplejson.dumps(
                    val,
                    default=encode_complex,
                    ignore_nan=True,
                    use_decimal=True,
                )
        elif column_type in STRING_SERIALIZABLE_COLUMN_TYPES:
            val = row[column]
            if val is not None:
                row[column] = str(val)

    return row


def cast_column_types(df: pd.DataFrame, column_types: Dict):
    for column, column_type in column_types.items():
        if column_type in CAST_TYPE_COLUMN_TYPES:
            try:
                df[column] = df[column].astype(column_type)
            except Exception:
                traceback.print_exc()
    return df


def deserialize_columns(row: pd.Series, column_types: Dict) -> pd.Series:
    for column, column_type in column_types.items():
        if column_type not in JSON_SERIALIZABLE_COLUMN_TYPES:
            continue

        val = row[column]
        if val is not None:
            row[column] = simplejson.loads(val)

    return row


def dask_from_pandas(df: pd.DataFrame) -> dd:
    ddf = dd.from_pandas(df, npartitions=1)
    npartitions = 1 + ddf.memory_usage(deep=True).sum().compute() // MAX_PARTITION_BYTE_SIZE
    ddf = ddf.repartition(npartitions=npartitions)

    return ddf


def apply_transform(ddf: dd, apply_function) -> dd:
    res = ddf.apply(apply_function, axis=1, meta=ddf)
    return res.compute()


def apply_transform_pandas(df: pd.DataFrame, apply_function) -> pd.DataFrame:
    return df.apply(apply_function, axis=1)



from mage_ai.api.resources.DatabaseResource import DatabaseResource
from mage_ai.orchestration.db.models.schedules import BlockRun, PipelineRun, PipelineSchedule
from mage_ai.orchestration.db import safe_db_query
from sqlalchemy.orm import aliased


class BlockRunResource(DatabaseResource):
    model_class = BlockRun

    @classmethod
    @safe_db_query
    def build_result_set(self, arr, user, **kwargs):
        block_runs = []
        for tup in arr:
            (
                block_uuid,
                completed_at,
                created_at,
                id,
                pipeline_run_id,
                status,
                updated_at,
                pipeline_schedule_id,
                pipeline_schedule_name,
            ) = tup

            block_run = BlockRun()
            block_run.block_uuid = block_uuid
            block_run.completed_at = completed_at
            block_run.created_at = created_at
            block_run.id = id
            block_run.pipeline_run_id = pipeline_run_id
            block_run.status = status
            block_run.updated_at = updated_at
            block_run.pipeline_schedule_id = pipeline_schedule_id
            block_run.pipeline_schedule_name = pipeline_schedule_name

            block_runs.append(block_run)

        return super().build_result_set(
            block_runs,
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def collection(self, query_arg, meta, user, **kwargs):
        pipeline_run = kwargs.get('parent_model')

        if pipeline_run:
            pipeline_run_id = pipeline_run.id

            return BlockRun.query.filter(BlockRun.pipeline_run_id == pipeline_run_id)

        query = BlockRun.query

        a = aliased(BlockRun, name='a')
        b = aliased(PipelineRun, name='b')
        c = aliased(PipelineSchedule, name='c')

        columns = [
            a.block_uuid,
            a.completed_at,
            a.created_at,
            a.id,
            a.pipeline_run_id,
            a.status,
            a.updated_at,
            c.id.label('pipeline_schedule_id'),
            c.name.label('pipeline_schedule_name'),
        ]

        query = (
            BlockRun.
            select(*columns).
            join(b, a.pipeline_run_id == b.id).
            join(c, b.pipeline_schedule_id == c.id)
        )

        pipeline_run_id = query_arg.get('pipeline_run_id', [None])
        if pipeline_run_id:
            pipeline_run_id = pipeline_run_id[0]
        if pipeline_run_id:
            query = (
                query.
                filter(a.pipeline_run_id == int(pipeline_run_id))
            )

        pipeline_uuid = query_arg.get('pipeline_uuid', [None])
        if pipeline_uuid:
            pipeline_uuid = pipeline_uuid[0]
        if pipeline_uuid:
            query = (
                query.
                filter(c.pipeline_uuid == pipeline_uuid)
            )

        return query.order_by(a.created_at.desc(), a.completed_at.desc())



from mage_ai.api.operations import constants
from mage_ai.api.presenters.BasePresenter import BasePresenter


class PipelineRunPresenter(BasePresenter):
    default_attributes = [
        'backfill_id',
        'completed_at',
        'created_at',
        'event_variables',
        'execution_date',
        'id',
        'metrics',
        'passed_sla',
        'pipeline_schedule_id',
        'pipeline_uuid',
        'status',
        'updated_at',
        'variables',
    ]

    async def present(self, **kwargs):
        if constants.LIST == kwargs['format']:
            return self.model.to_dict(include_attributes=[
                'block_runs',
                'block_runs_count',
                'pipeline_schedule_name',
                'pipeline_schedule_token',
                'pipeline_schedule_type',
            ])
        elif constants.DETAIL == kwargs['format']:
            block_runs = self.model.block_runs
            data = self.model.to_dict()
            pipeline_schedule = self.resource.pipeline_schedule

            arr = []
            for r in block_runs:
                block_run = r.to_dict()
                block_run['pipeline_schedule_id'] = pipeline_schedule.id
                block_run['pipeline_schedule_name'] = pipeline_schedule.name
                arr.append(block_run)
            arr.sort(key=lambda b: b.get('created_at'))
            data['block_runs'] = arr
            return data

        return self.model.to_dict()


PipelineRunPresenter.register_format(
    constants.LIST,
    PipelineRunPresenter.default_attributes + [
        'block_runs',
        'block_runs_count',
        'pipeline_schedule_name',
        'pipeline_schedule_token',
        'pipeline_schedule_type',
    ],
)

PipelineRunPresenter.register_format(
    constants.DETAIL,
    PipelineRunPresenter.default_attributes + [
        'block_runs',
    ],
)



from mage_ai.api.presenters.BasePresenter import BasePresenter


class GitBranchPresenter(BasePresenter):
    default_attributes = [
        'name',
        'status',
        'message',
        'action_type',
        'files',
    ]



from datetime import datetime
from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.block.sql import (
    bigquery,
    clickhouse,
    mssql,
    mysql,
    postgres,
    redshift,
    snowflake,
    trino,
)
from mage_ai.data_preparation.models.block.sql.utils.shared import (
    has_create_or_insert_statement,
    interpolate_vars,
)
from mage_ai.data_preparation.models.constants import BlockType
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.base import (
    DataSource,
    ExportWritePolicy,
    QUERY_ROW_LIMIT,
)
from mage_ai.io.config import ConfigFileLoader
from os import path
from time import sleep
from typing import Any, Dict, List
import re

MAGE_SEMI_COLON = '__MAGE_SEMI_COLON__'
PREVIEWABLE_BLOCK_TYPES = [
    BlockType.DATA_EXPORTER,
    BlockType.DATA_LOADER,
    BlockType.DBT,
    BlockType.TRANSFORMER,
]


def execute_sql_code(
    block,
    query: str,
    execution_partition: str = None,
    global_vars: Dict = None,
    config_file_loader: Any = None,
    configuration: Dict = None,
    test_execution: bool = False,
) -> List[Any]:
    configuration = configuration if configuration else block.configuration
    use_raw_sql = configuration.get('use_raw_sql')

    if not config_file_loader:
        config_path = path.join(get_repo_path(), 'io_config.yaml')
        config_profile = configuration.get('data_provider_profile')
        config_file_loader = ConfigFileLoader(config_path, config_profile)

    data_provider = configuration.get('data_provider')
    database = configuration.get('data_provider_database')
    schema = configuration.get('data_provider_schema')
    export_write_policy = configuration.get('export_write_policy', ExportWritePolicy.APPEND)

    if 'execution_date' in global_vars:
        global_vars['ds'] = global_vars['execution_date'].strftime('%Y-%m-%d')

    block.set_global_vars(global_vars)

    table_name = block.table_name
    should_query = block.type in PREVIEWABLE_BLOCK_TYPES

    kwargs_shared = dict(
        drop_table_on_replace=True,
        if_exists=export_write_policy,
        index=False,
        verbose=BlockType.DATA_EXPORTER == block.type,
    )

    limit = int(configuration.get('limit') or QUERY_ROW_LIMIT)
    if test_execution:
        limit = min(limit, QUERY_ROW_LIMIT)
    else:
        limit = QUERY_ROW_LIMIT

    if DataSource.BIGQUERY.value == data_provider:
        from mage_ai.io.bigquery import BigQuery

        loader = BigQuery.with_config(config_file_loader)
        database = database or loader.default_database()

        bigquery.create_upstream_block_tables(
            loader,
            block,
            configuration=configuration,
            execution_partition=execution_partition,
            query=query,
        )

        query_string = bigquery.interpolate_input_data(block, query, loader)
        query_string = interpolate_vars(query_string, global_vars=global_vars)

        if use_raw_sql:
            return execute_raw_sql(
                loader,
                block,
                query_string,
                configuration=configuration,
                should_query=should_query,
            )
        else:
            loader.export(
                None,
                f'{schema}.{table_name}',
                database=database,
                if_exists=export_write_policy,
                query_string=query_string,
                verbose=BlockType.DATA_EXPORTER == block.type,
            )

            if should_query:
                """
                An error is thrown because the table doesn’t exist until you re-run the query
                NotFound: 404 Not found: Table database:schema.table_name
                    was not found in location XX
                """
                tries = 0
                while tries < 10:
                    sleep(tries)
                    tries += 1
                    try:
                        result = loader.load(
                            f'SELECT * FROM {database}.{schema}.{table_name}',
                            limit=limit,
                            verbose=False,
                        )
                        return [result]
                    except Exception as err:
                        if '404' not in str(err):
                            raise err
    elif DataSource.CLICKHOUSE.value == data_provider:
        from mage_ai.io.clickhouse import ClickHouse

        loader = ClickHouse.with_config(config_file_loader)
        clickhouse.create_upstream_block_tables(
            loader,
            block,
            configuration=configuration,
            execution_partition=execution_partition,
            query=query,
        )

        query_string = clickhouse.interpolate_input_data(block, query)
        query_string = interpolate_vars(
            query_string, global_vars=global_vars)

        database = database or 'default'

        if use_raw_sql:
            return execute_raw_sql(
                loader,
                block,
                query_string,
                configuration=configuration,
                should_query=should_query,
            )
        else:
            loader.export(
                None,
                table_name=table_name,
                database=database,
                query_string=query_string,
                **kwargs_shared,
            )

            if should_query:
                return [
                    loader.load(
                        f'SELECT * FROM {database}.{table_name}',
                        verbose=False,
                    ),
                ]
    elif DataSource.MSSQL.value == data_provider:
        from mage_ai.io.mssql import MSSQL

        with MSSQL.with_config(config_file_loader) as loader:
            mssql.create_upstream_block_tables(
                loader,
                block,
                configuration=configuration,
                execution_partition=execution_partition,
                query=query,
            )

            query_string = mssql.interpolate_input_data(block, query)
            query_string = interpolate_vars(query_string, global_vars=global_vars)

            if use_raw_sql:
                return execute_raw_sql(
                    loader,
                    block,
                    query_string,
                    configuration=configuration,
                    should_query=should_query,
                )
            else:
                loader.export(
                    None,
                    None,
                    table_name,
                    query_string=query_string,
                    drop_table_on_replace=True,
                    if_exists=export_write_policy,
                    index=False,
                    verbose=BlockType.DATA_EXPORTER == block.type,
                )

                if should_query:
                    return [
                        loader.load(
                            f'SELECT * FROM {table_name}',
                            limit=limit,
                            verbose=False,
                        ),
                    ]
    elif DataSource.MYSQL.value == data_provider:
        from mage_ai.io.mysql import MySQL

        with MySQL.with_config(config_file_loader) as loader:
            mysql.create_upstream_block_tables(
                loader,
                block,
                configuration=configuration,
                execution_partition=execution_partition,
                query=query,
            )

            query_string = mysql.interpolate_input_data(block, query)
            query_string = interpolate_vars(query_string, global_vars=global_vars)

            if use_raw_sql:
                return execute_raw_sql(
                    loader,
                    block,
                    query_string,
                    configuration=configuration,
                    should_query=should_query,
                )
            else:
                loader.export(
                    None,
                    None,
                    table_name,
                    query_string=query_string,
                    **kwargs_shared,
                )

                if should_query:
                    return [
                        loader.load(
                            f'SELECT * FROM {table_name}',
                            limit=limit,
                            verbose=False,
                        ),
                    ]
    elif DataSource.POSTGRES.value == data_provider:
        from mage_ai.io.postgres import Postgres

        with Postgres.with_config(config_file_loader) as loader:
            postgres.create_upstream_block_tables(
                loader,
                block,
                configuration=configuration,
                execution_partition=execution_partition,
                query=query,
            )

            query_string = postgres.interpolate_input_data(block, query, loader)
            query_string = interpolate_vars(query_string, global_vars=global_vars)

            schema = schema or loader.default_schema()

            if use_raw_sql:
                return execute_raw_sql(
                    loader,
                    block,
                    query_string,
                    configuration=configuration,
                    should_query=should_query,
                )
            else:
                loader.export(
                    None,
                    schema,
                    table_name,
                    query_string=query_string,
                    **kwargs_shared,
                )

                if should_query:
                    return [
                        loader.load(
                            f'SELECT * FROM {schema}.{table_name}',
                            limit=limit,
                            verbose=False,
                        ),
                    ]
    elif DataSource.REDSHIFT.value == data_provider:
        from mage_ai.io.redshift import Redshift

        with Redshift.with_config(config_file_loader) as loader:
            redshift.create_upstream_block_tables(
                loader,
                block,
                configuration=configuration,
                execution_partition=execution_partition,
                query=query,
            )

            database = database or loader.default_database()
            schema = schema or loader.default_schema()

            query_string = redshift.interpolate_input_data(block, query, loader)
            query_string = interpolate_vars(query_string, global_vars=global_vars)

            if use_raw_sql:
                return execute_raw_sql(
                    loader,
                    block,
                    query_string,
                    configuration=configuration,
                    should_query=should_query,
                )
            else:
                loader.export(
                    None,
                    table_name,
                    schema=schema,
                    query_string=query_string,
                    **kwargs_shared,
                )

                if should_query:
                    return [
                            loader.load(
                                f'SELECT * FROM {schema}.{table_name}',
                                limit=limit,
                                verbose=False,
                            ),
                        ]
    elif DataSource.SNOWFLAKE.value == data_provider:
        from mage_ai.io.snowflake import Snowflake

        table_name = table_name.upper() if table_name else table_name

        with Snowflake.with_config(config_file_loader, database=database, schema=schema) as loader:
            database = database or loader.default_database()
            database = database.upper() if database else database

            schema = schema or loader.default_schema()
            schema = schema.upper() if schema else schema

            snowflake.create_upstream_block_tables(
                loader,
                block,
                configuration=configuration,
                execution_partition=execution_partition,
                query=query,
            )

            query_string = snowflake.interpolate_input_data(block, query, loader)
            query_string = interpolate_vars(query_string, global_vars=global_vars)

            if use_raw_sql:
                return execute_raw_sql(
                    loader,
                    block,
                    query_string,
                    configuration=configuration,
                    should_query=should_query,
                )
            else:
                loader.export(
                    None,
                    table_name,
                    database,
                    schema,
                    if_exists=export_write_policy,
                    query_string=query_string,
                    verbose=BlockType.DATA_EXPORTER == block.type,
                )

                if should_query:
                    return [
                        loader.load(
                            f'SELECT * FROM "{database}"."{schema}"."{table_name}"',
                            database=database,
                            schema=schema,
                            table_name=table_name,
                            limit=limit,
                            verbose=False,
                        ),
                    ]
    elif DataSource.TRINO.value == data_provider:
        from mage_ai.io.trino import Trino

        unique_table_name_suffix = None
        if (configuration or block.configuration).get('unique_upstream_table_name', False):
            unique_table_name_suffix = str(int(datetime.utcnow().timestamp()))

        with Trino.with_config(config_file_loader) as loader:
            database = database or loader.default_database()
            schema = schema or loader.default_schema()

            trino.create_upstream_block_tables(
                loader,
                block,
                configuration=configuration,
                execution_partition=execution_partition,
                query=query,
                unique_table_name_suffix=unique_table_name_suffix,
            )

            query_string = trino.interpolate_input_data(
                block,
                query,
                loader,
                unique_table_name_suffix=unique_table_name_suffix,
            )
            query_string = interpolate_vars(query_string, global_vars=global_vars)

            if use_raw_sql:
                return execute_raw_sql(
                    loader,
                    block,
                    query_string,
                    configuration=configuration,
                    should_query=should_query,
                )
            else:
                loader.export(
                    None,
                    schema,
                    table_name,
                    drop_table_on_replace=True,
                    if_exists=export_write_policy,
                    query_string=query_string,
                    verbose=BlockType.DATA_EXPORTER == block.type,
                )

                if should_query:
                    names = list(filter(lambda x: x, [
                        database,
                        schema,
                        table_name,
                    ]))
                    full_table_name = '.'.join([f'"{n}"' for n in names])

                    return [
                        loader.load(
                            f'SELECT * FROM {full_table_name}',
                            limit=limit,
                            verbose=False,
                        ),
                    ]


def split_query_string(query_string: str) -> List[str]:
    text_parts = []

    matches = re.finditer(r"'(.*?)'|\"(.*?)\"", query_string, re.IGNORECASE)

    previous_idx = 0

    for idx, match in enumerate(matches):
        matched_string = match.group()
        updated_string = re.sub(r';', MAGE_SEMI_COLON, matched_string)

        start_idx, end_idx = match.span()

        previous_chunk = query_string[previous_idx:start_idx]
        text_parts.append(previous_chunk)
        text_parts.append(updated_string)
        previous_idx = end_idx

    text_parts.append(query_string[previous_idx:])

    text_combined = ''.join(text_parts)
    queries = text_combined.split(';')

    arr = []
    for query in queries:
        query = query.strip()
        if not query:
            continue

        lines = query.split('\n')
        query = '\n'.join(list(filter(lambda x: not x.startswith('--'), lines)))
        query = query.strip()
        query = re.sub(MAGE_SEMI_COLON, ';', query)

        if query:
            arr.append(query)

    return arr


def execute_raw_sql(
    loader,
    block: 'Block',
    query_string: str,
    configuration: Dict = {},
    should_query: bool = False,
) -> List[Any]:
    queries = []
    fetch_query_at_indexes = []

    has_create_or_insert = has_create_or_insert_statement(query_string)

    for query in split_query_string(query_string):
        if has_create_or_insert:
            queries.append(query)
            fetch_query_at_indexes.append(False)
        else:
            queries.append(query)
            fetch_query_at_indexes.append(True)

    if should_query and has_create_or_insert:
        queries.append(f'SELECT * FROM {block.full_table_name} LIMIT 1000')
        fetch_query_at_indexes.append(block.full_table_name)

    results = loader.execute_queries(
        queries,
        commit=True,
        fetch_query_at_indexes=fetch_query_at_indexes,
    )

    if should_query:
        return [results[-1]]

    return []


class SQLBlock(Block):
    def _execute_block(
        self,
        outputs_from_input_vars,
        custom_code=None,
        execution_partition=None,
        global_vars=None,
        **kwargs,
    ) -> List:
        test_execution = kwargs.get('test_execution')
        return execute_sql_code(
            self,
            custom_code or self.content,
            execution_partition=execution_partition,
            global_vars=global_vars,
            test_execution=test_execution,
        )



from mage_ai.data_preparation.logging.logger import DictLogger
from mage_ai.data_preparation.logging.logger_manager_factory import LoggerManagerFactory
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.shared.hash import merge_dict
from typing import Dict
import asyncio


class PipelineExecutor:
    def __init__(self, pipeline: Pipeline, execution_partition: str = None):
        self.pipeline = pipeline
        self.execution_partition = execution_partition
        self.logger_manager = LoggerManagerFactory.get_logger_manager(
            pipeline_uuid=self.pipeline.uuid,
            partition=self.execution_partition,
            repo_config=self.pipeline.repo_config,
        )
        self.logger = DictLogger(self.logger_manager.logger)

    def execute(
        self,
        analyze_outputs: bool = False,
        global_vars: Dict = None,
        run_sensors: bool = True,
        run_tests: bool = True,
        update_status: bool = False,
        **kwargs,
    ) -> None:
        asyncio.run(self.pipeline.execute(
            analyze_outputs=analyze_outputs,
            global_vars=global_vars,
            run_sensors=run_sensors,
            run_tests=run_tests,
            update_status=update_status,
        ))
        self.logger_manager.output_logs_to_destination()

    def _build_tags(self, **kwargs):
        default_tags = dict(
            pipeline_uuid=self.pipeline.uuid,
        )
        if kwargs.get('pipeline_run_id'):
            default_tags['pipeline_run_id'] = kwargs.get('pipeline_run_id')
        return merge_dict(kwargs.get('tags', {}), default_tags)



from mage_ai.data_preparation.logging.logger import DictLogger
from mage_ai.data_preparation.logging.logger_manager_factory import LoggerManagerFactory
from mage_ai.data_preparation.models.block.dbt.utils import run_dbt_tests
from mage_ai.data_preparation.models.constants import BlockType, PipelineType
from mage_ai.orchestration.db.models.schedules import PipelineRun
from mage_ai.shared.hash import merge_dict
from mage_ai.shared.utils import clean_name
from typing import Callable, Dict, List, Union
import json
import requests
import traceback


class BlockExecutor:
    def __init__(
        self,
        pipeline,
        block_uuid,
        execution_partition=None
    ):
        self.pipeline = pipeline
        self.block_uuid = block_uuid
        self.block = self.pipeline.get_block(self.block_uuid, check_template=True)
        self.execution_partition = execution_partition
        self.logger_manager = LoggerManagerFactory.get_logger_manager(
            pipeline_uuid=self.pipeline.uuid,
            block_uuid=clean_name(self.block_uuid),
            partition=self.execution_partition,
            repo_config=self.pipeline.repo_config,
        )
        self.logger = DictLogger(self.logger_manager.logger)

    def execute(
        self,
        analyze_outputs: bool = False,
        callback_url: Union[str, None] = None,
        global_vars: Union[Dict, None] = None,
        update_status: bool = False,
        on_complete: Union[Callable[[str], None], None] = None,
        on_failure: Union[Callable[[str, Dict], None], None] = None,
        on_start: Union[Callable[[str], None], None] = None,
        input_from_output: Union[Dict, None] = None,
        verify_output: bool = True,
        runtime_arguments: Union[Dict, None] = None,
        template_runtime_configuration: Union[Dict, None] = None,
        dynamic_block_index: Union[int, None] = None,
        dynamic_block_uuid: Union[str, None] = None,
        dynamic_upstream_block_uuids: Union[List[str], None] = None,
        **kwargs,
    ) -> Dict:
        if template_runtime_configuration:
            self.block.template_runtime_configuration = template_runtime_configuration
        try:
            result = dict()

            tags = self._build_tags(**kwargs)

            self.logger.info(f'Start executing block with {self.__class__.__name__}.', **tags)
            if on_start is not None:
                on_start(self.block_uuid)
            pipeline_run = PipelineRun.query.get(kwargs['pipeline_run_id']) \
                if 'pipeline_run_id' in kwargs else None
            try:
                result = self._execute(
                    analyze_outputs=analyze_outputs,
                    callback_url=callback_url,
                    global_vars=global_vars,
                    update_status=update_status,
                    input_from_output=input_from_output,
                    logging_tags=tags,
                    verify_output=verify_output,
                    runtime_arguments=runtime_arguments,
                    template_runtime_configuration=template_runtime_configuration,
                    dynamic_block_index=dynamic_block_index,
                    dynamic_block_uuid=dynamic_block_uuid,
                    dynamic_upstream_block_uuids=dynamic_upstream_block_uuids,
                    **kwargs,
                )
            except Exception as e:
                self.logger.exception(
                    f'Failed to execute block {self.block.uuid}',
                    **merge_dict(tags, dict(
                        block_type=self.block.type,
                        block_uuid=self.block.uuid,
                        error=e
                    )),
                )
                if on_failure is not None:
                    on_failure(
                        self.block_uuid,
                        error=dict(
                            error=str(e),
                            errors=traceback.format_stack(),
                            message=traceback.format_exc(),
                        ),
                    )
                else:
                    self.__update_block_run_status(
                        'failed',
                        block_run_id=kwargs.get('block_run_id'),
                        callback_url=callback_url,
                        tags=tags,
                    )
                self._execute_callback(
                    'on_failure',
                    global_vars=global_vars,
                    logging_tags=tags,
                    pipeline_run=pipeline_run,
                )
                raise e
            self.logger.info(f'Finish executing block with {self.__class__.__name__}.', **tags)
            if on_complete is not None:
                on_complete(self.block_uuid)
            else:
                self.__update_block_run_status(
                    'completed',
                    block_run_id=kwargs.get('block_run_id'),
                    callback_url=callback_url,
                    tags=tags
                )
            self._execute_callback(
                'on_success',
                global_vars=global_vars,
                logging_tags=tags,
                pipeline_run=pipeline_run,
            )

            return result
        finally:
            self.logger_manager.output_logs_to_destination()

    def _execute(
        self,
        analyze_outputs: bool = False,
        callback_url: Union[str, None] = None,
        global_vars: Union[Dict, None] = None,
        update_status: bool = False,
        input_from_output: Union[Dict, None] = None,
        logging_tags: Dict = dict(),
        verify_output: bool = True,
        runtime_arguments: Union[Dict, None] = None,
        dynamic_block_index: Union[int, None] = None,
        dynamic_block_uuid: Union[str, None] = None,
        dynamic_upstream_block_uuids: Union[List[str], None] = None,
        **kwargs,
    ) -> Dict:
        result = self.block.execute_sync(
            analyze_outputs=analyze_outputs,
            execution_partition=self.execution_partition,
            global_vars=global_vars,
            logger=self.logger,
            logging_tags=logging_tags,
            run_all_blocks=True,
            update_status=update_status,
            input_from_output=input_from_output,
            verify_output=verify_output,
            runtime_arguments=runtime_arguments,
            dynamic_block_index=dynamic_block_index,
            dynamic_block_uuid=dynamic_block_uuid,
            dynamic_upstream_block_uuids=dynamic_upstream_block_uuids,
        )

        if BlockType.DBT == self.block.type:
            run_dbt_tests(
                block=self.block,
                global_vars=global_vars,
                logger=self.logger,
                logging_tags=logging_tags,
            )
        elif PipelineType.INTEGRATION != self.pipeline.type:
            self.block.run_tests(
                execution_partition=self.execution_partition,
                global_vars=global_vars,
                logger=self.logger,
                logging_tags=logging_tags,
                update_tests=False,
                dynamic_block_uuid=dynamic_block_uuid,
            )

        return result

    def _execute_callback(
        self,
        callback: str,
        global_vars,
        logging_tags,
        pipeline_run,
    ):
        if self.block.callback_block:
            try:
                self.block.callback_block.execute_callback(
                    callback,
                    global_vars=global_vars,
                    logger=self.logger,
                    logging_tags=logging_tags,
                    pipeline_run=pipeline_run,
                )
            except Exception as callback_err:
                self.logger.exception(
                    f'Failed to execute {callback} callback block for {self.block.uuid}',
                    **merge_dict(logging_tags, dict(
                        error=callback_err,
                    )),
                )

    def _run_commands(
        self,
        block_run_id: int = None,
        global_vars: Dict = None,
        **kwargs,
    ) -> List[str]:
        cmd = f'/app/run_app.sh '\
              f'mage run {self.pipeline.repo_config.repo_path} {self.pipeline.uuid}'
        options = [
            '--block-uuid',
            self.block_uuid,
            '--executor-type',
            'local_python',
        ]
        if self.execution_partition is not None:
            options += ['--execution-partition', self.execution_partition]
        if block_run_id is not None:
            options += ['--block-run-id', f'{block_run_id}']
        if kwargs.get('pipeline_run_id'):
            pipeline_run_id = kwargs.get('pipeline_run_id')
            options += [
                '--pipeline-run-id',
                f'{pipeline_run_id}',
            ]
        if kwargs.get('template_runtime_configuration'):
            template_run_configuration = kwargs.get('template_runtime_configuration')
            options += [
                '--template-runtime-configuration',
                json.dumps(template_run_configuration),
            ]
        return cmd.split(' ') + options

    def __update_block_run_status(
        self,
        status: str,
        block_run_id: int = None,
        callback_url: str = None,
        tags: Dict = dict(),
    ):
        """
        Update the status of block run by edither updating the BlockRun db object or making
        API call

        Args:
            status (str): 'completed' or 'failed'
            block_run_id (int): the id of the block run
            callback_url (str): with format http(s)://[host]:[port]/api/block_runs/[block_run_id]
            tags (dict): tags used in logging
        """
        if not block_run_id and not callback_url:
            return
        try:
            if not block_run_id:
                block_run_id = int(callback_url.split('/')[-1])

            from mage_ai.orchestration.db.models.schedules import BlockRun

            block_run = BlockRun.query.get(block_run_id)
            block_run.update(status=status)
            return
        except Exception:
            pass

        # Fall back to making API calls
        response = requests.put(
            callback_url,
            data=json.dumps({
                'block_run': {
                    'status': status,
                },
            }),
            headers={
                'Content-Type': 'application/json',
            },
        )
        self.logger.info(
            f'Callback response: {response.text}',
            **tags,
        )

    def _build_tags(self, **kwargs):
        default_tags = dict(
            block_type=self.block.type,
            block_uuid=self.block_uuid,
            pipeline_uuid=self.pipeline.uuid,
        )
        if kwargs.get('block_run_id'):
            default_tags['block_run_id'] = kwargs.get('block_run_id')
        if kwargs.get('pipeline_run_id'):
            default_tags['pipeline_run_id'] = kwargs.get('pipeline_run_id')
        return merge_dict(kwargs.get('tags', {}), default_tags)



from mage_ai.orchestration.queue.config import QueueConfig
from mage_ai.orchestration.queue.process_queue import JobStatus, ProcessQueue
from mage_ai.tests.base_test import TestCase


class ProcessQueueTests(TestCase):
    def setUp(self):
        queue_config = QueueConfig.load(config=dict(concurrency=100))
        self.queue = ProcessQueue(queue_config=queue_config)

    def test_init(self):
        self.assertEqual(self.queue.size, 100)

    def test_clean_up_jobs(self):
        self.queue.job_dict['block_run_1'] = JobStatus.QUEUED
        self.queue.job_dict['block_run_2'] = 100
        self.queue.job_dict['block_run_3'] = JobStatus.COMPLETED
        self.queue.job_dict['block_run_4'] = JobStatus.CANCELLED
        self.queue.clean_up_jobs()
        self.assertEqual(self.queue.job_dict['block_run_1'], JobStatus.QUEUED)
        self.assertEqual(self.queue.job_dict['block_run_2'], 100)
        self.assertFalse('block_run_3' in self.queue.job_dict)
        self.assertFalse('block_run_4' in self.queue.job_dict)

    def test_has_job(self):
        self.queue.job_dict['block_run_1'] = JobStatus.QUEUED
        self.queue.job_dict['block_run_2'] = 100
        self.queue.job_dict['block_run_3'] = JobStatus.COMPLETED
        self.queue.job_dict['block_run_4'] = JobStatus.CANCELLED
        self.assertTrue(self.queue.has_job('block_run_1'))
        self.assertTrue(self.queue.has_job('block_run_2'))
        self.assertFalse(self.queue.has_job('block_run_3'))
        self.assertFalse(self.queue.has_job('block_run_4'))
        self.assertFalse(self.queue.has_job('block_run_5'))



from mage_ai.orchestration.triggers.event_trigger import EventTrigger
from mage_ai.tests.base_test import TestCase
from unittest.mock import patch


class EventTriggerTests(TestCase):
    @patch('mage_ai.orchestration.triggers.event_trigger.schedule_with_event')
    def test_run(self, mock_schedule):
        test_event = dict(a=1, b=2)
        trigger = EventTrigger()
        trigger.run(test_event)
        mock_schedule.assert_called_once_with(test_event)



from mage_ai.streaming.sources.azure_event_hub import AzureEventHubSource
from mage_ai.streaming.sources.kafka import KafkaSource
from mage_ai.streaming.sources.rabbitmq import RabbitMQSource
from mage_ai.streaming.sources.source_factory import SourceFactory
from mage_ai.tests.base_test import TestCase
from unittest.mock import patch


class SourceFactoryTests(TestCase):
    def test_get_source_kafka(self):
        with patch.object(KafkaSource,
                          '__init__',
                          return_value=None) as mock_init:
            config = dict(
                connector_type='kafka',
            )
            source = SourceFactory.get_source(config)
            self.assertIsInstance(source, KafkaSource)
            mock_init.assert_called_once_with(config)

    def test_get_source_azure_event_hub(self):
        with patch.object(AzureEventHubSource,
                          '__init__',
                          return_value=None) as mock_init:
            config = dict(
                connector_type='azure_event_hub',
            )
            source = SourceFactory.get_source(config)
            self.assertIsInstance(source, AzureEventHubSource)
            mock_init.assert_called_once_with(config)

    def test_get_source_rabbitmq(self):
        with patch.object(RabbitMQSource,
                          '__init__',
                          return_value=None) as mock_init:
            config = dict(
                connector_type='rabbitmq',
            )
            source = SourceFactory.get_source(config)
            self.assertIsInstance(source, RabbitMQSource)
            mock_init.assert_called_once_with(config)

    def test_get_source_other(self):
        s = "Consuming data from random is not supported " \
            "in streaming pipelines yet."
        with self.assertRaises(Exception) as context:
            SourceFactory.get_source(dict(
                connector_type='random',
            ))
        self.assertTrue(
            s
            in str(context.exception),
        )



from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig
import os

GIT_SSH_PRIVATE_KEY_SECRET_NAME = 'mage_git_ssh_private_key_b64'
GIT_SSH_PUBLIC_KEY_SECRET_NAME = 'mage_git_ssh_public_key_b64'


@dataclass
class GitConfig(BaseConfig):
    remote_repo_link: str
    repo_path: str = os.getcwd()
    username: str = ''
    email: str = ''
    branch: str = 'main'
    sync_on_pipeline_run: bool = False
    ssh_private_key_secret_name: str = GIT_SSH_PRIVATE_KEY_SECRET_NAME
    ssh_public_key_secret_name: str = GIT_SSH_PUBLIC_KEY_SECRET_NAME
    # This is not necessary anymore, but leaving it for backwards compatibility
    type: str = 'git'



from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis
from mage_ai.data_cleaner.transformer_actions.utils import columns_to_remove
from mage_ai.tests.base_test import TestCase


class UtilsTests(TestCase):
    def test_columns_to_remove(self):
        transformer_actions = [
            dict(
                action_type=ActionType.FILTER,
                axis=Axis.COLUMN,
                action_arguments=['wand'],
            ),
            dict(
                action_type=ActionType.REMOVE,
                axis=Axis.ROW,
                action_arguments=['spear'],
            ),
            dict(
                action_type=ActionType.REMOVE,
                axis=Axis.COLUMN,
                action_arguments=['sword'],
            ),
        ]
        self.assertEqual(columns_to_remove(transformer_actions), ['sword'])



from mage_ai.data_preparation.models.block import (
    Block,
    CallbackBlock,
    SensorBlock,
)
from mage_ai.data_preparation.models.block.dbt import DBTBlock
from mage_ai.data_preparation.models.block.extension.block import ExtensionBlock
from mage_ai.data_preparation.models.constants import BlockType


BLOCK_TYPE_TO_CLASS = {
    BlockType.CALLBACK: CallbackBlock,
    BlockType.CUSTOM: Block,
    BlockType.DATA_EXPORTER: Block,
    BlockType.DATA_LOADER: Block,
    BlockType.DBT: DBTBlock,
    BlockType.EXTENSION: ExtensionBlock,
    BlockType.SCRATCHPAD: Block,
    BlockType.TRANSFORMER: Block,
    BlockType.SENSOR: SensorBlock,
}

TAG_DYNAMIC = 'dynamic'
TAG_DYNAMIC_CHILD = 'dynamic_child'
TAG_REDUCE_OUTPUT = 'reduce_output'



from datetime import datetime
from mage_ai.orchestration.db import db_connection, safe_db_query
from mage_ai.shared.strings import camel_to_snake_case
from sqlalchemy import (
    Column,
    DateTime,
    Integer,
)
from sqlalchemy.ext.declarative import declared_attr, declarative_base
from sqlalchemy.orm.collections import InstrumentedList
from sqlalchemy.sql import func
from typing import Dict

Base = declarative_base()


class classproperty(property):
    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


class BaseModel(Base):
    __abstract__ = True

    @declared_attr
    def __tablename__(cls):
        return camel_to_snake_case(cls.__name__)

    @classproperty
    def query(cls):
        return db_connection.session.query(cls)

    @classproperty
    def select(cls):
        return db_connection.session.query

    @property
    def session(self):
        return db_connection.session

    id = Column(
        Integer,
        primary_key=True,
        autoincrement=True,
    )
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
    )

    @classmethod
    def create(self, **kwargs):
        commit = kwargs.get('commit', True)
        kwargs.pop('commit', None)
        model = self(**kwargs)
        model.save(commit=commit)
        return model

    def full_clean(self, **kwargs) -> None:
        pass

    @classmethod
    @safe_db_query
    def get(self, uuid):
        return self.query.get(uuid)

    def save(self, commit=True) -> None:
        self.session.add(self)
        if commit:
            try:
                self.session.commit()
            except Exception as e:
                self.session.rollback()
                raise e

    def update(self, **kwargs) -> None:
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
        try:
            self.session.commit()
        except Exception as e:
            self.session.rollback()
            raise e

    def delete(self, commit: bool = True) -> None:
        self.session.delete(self)
        if commit:
            try:
                self.session.commit()
            except Exception as e:
                self.session.rollback()
                raise e

    def refresh(self):
        self.session.refresh(self)

    def to_dict(self, include_attributes=[]) -> Dict:
        def __format_value(value):
            if type(value) is datetime:
                return str(value)
            elif type(value) is InstrumentedList:
                try:
                    value = sorted(value, key=lambda x: x.id)
                except Exception:
                    pass
                return [__format_value(v) for v in value]
            elif hasattr(value, 'to_dict'):
                return value.to_dict()
            return value
        obj_dict = {
            c.name: __format_value(getattr(self, c.name))
            for c in self.__table__.columns
        }
        if include_attributes is not None and len(include_attributes) > 0:
            for attr in include_attributes:
                if hasattr(self, attr):
                    obj_dict[attr] = __format_value(getattr(self, attr))
        return obj_dict



from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.bigquery import BigQuery
from mage_ai.io.config import ConfigFileLoader
from pandas import DataFrame
from os import path

if 'data_exporter' not in globals():
    from mage_ai.data_preparation.decorators import data_exporter


@data_exporter
def export_data_to_big_query(df: DataFrame, **kwargs) -> None:
    """
    Template for exporting data to a BigQuery warehouse.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://docs.mage.ai/design/data-loading#bigquery
    """
    table_id = 'your-project.your_dataset.your_table_name'
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    BigQuery.with_config(ConfigFileLoader(config_path, config_profile)).export(
        df,
        table_id,
        if_exists='replace',  # Specify resolution policy if table name already exists
    )



from faker import Faker
from mage_ai.api.operations import constants
from mage_ai.api.operations.base import BaseOperation
from mage_ai.orchestration.db.models.oauth import User
from mage_ai.shared.array import find
from mage_ai.tests.base_test import AsyncDBTestCase as TestCase
from typing import Dict, List, Union
import inflection


class BaseApiTestCase(TestCase):
    model_class = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.faker = Faker()

    @property
    def model_class_name(self) -> str:
        return self.model_class.__name__.lower()

    @property
    def model_class_name_plural(self) -> str:
        return inflection.pluralize(self.model_class_name)

    def build_operation(self, **kwargs) -> BaseOperation:
        user = None
        if 'user' in kwargs:
            user = kwargs['user']
        else:
            user = User(owner=True)

        return BaseOperation(
            action=kwargs.get('action'),
            meta=kwargs.get('meta', {}),
            options=kwargs.get('options', {}),
            payload=kwargs.get('payload', {}),
            pk=kwargs.get('pk'),
            query=kwargs.get('query', {}),
            resource=kwargs.get('resource'),
            resource_parent=kwargs.get('resource_parent'),
            resource_parent_id=kwargs.get('resource_parent_id'),
            user=user,
        )

    def build_create_operation(self, payload: Dict, **kwargs) -> BaseOperation:
        return self.build_operation(
            action=constants.CREATE,
            payload={
                self.model_class_name: payload,
            },
            resource=self.model_class_name_plural,
            **kwargs,
        )

    def build_delete_operation(self, pk: Union[int, str], **kwargs) -> BaseOperation:
        return self.build_operation(
            action=constants.DELETE,
            pk=pk,
            resource=self.model_class_name_plural,
            **kwargs,
        )

    def build_detail_operation(self, pk: Union[int, str], **kwargs) -> BaseOperation:
        return self.build_operation(
            action=constants.DETAIL,
            pk=pk,
            resource=self.model_class_name_plural,
            **kwargs,
        )

    def build_list_operation(self, **kwargs) -> BaseOperation:
        return self.build_operation(
            action=constants.LIST,
            resource=self.model_class_name_plural,
            **kwargs,
        )

    def build_update_operation(self, pk: Union[int, str], payload: Dict, **kwargs) -> BaseOperation:
        return self.build_operation(
            action=constants.UPDATE,
            payload=payload,
            pk=pk,
            resource=self.model_class_name_plural,
            **kwargs,
        )

    async def base_test_execute_create(
        self,
        payload: Dict,
        after_create_count: int = 1,
        before_create_count: int = 0,
        **kwargs,
    ) -> Dict:
        self.assertEqual(self.model_class.query.count(), before_create_count)

        operation = self.build_create_operation(payload, **kwargs)
        response = await operation.execute()

        if 'error' in response:
            raise Exception(response['error'])

        self.assertEqual(self.model_class.query.count(), after_create_count)

        return response

    async def base_test_execute_delete(
        self,
        pk,
        **kwargs,
    ) -> Dict:
        operation = self.build_delete_operation(pk, **kwargs)
        response = await operation.execute()

        if 'error' in response:
            raise Exception(response['error'])

        self.assertIsNone(self.model_class.query.get(pk))

        return response

    async def base_test_execute_detail(
        self,
        pk,
        model_dict: Dict,
        **kwargs,
    ) -> Dict:
        operation = self.build_detail_operation(pk, **kwargs)
        response = await operation.execute()

        if 'error' in response:
            raise Exception(response['error'])

        for key, value in model_dict.items():
            self.assertEqual(response[self.model_class_name][key], value)

        return response

    async def base_test_execute_list(
        self,
        create_payloads: List[Dict] = [{}],
        model_fields_to_check: List[str] = [],
        **kwargs,
    ) -> Dict:
        models = []
        for payload in create_payloads:
            response = await self.build_create_operation(payload, **kwargs).execute()
            if 'error' in response:
                raise Exception(response['error'])
            models.append(response[self.model_class_name])

        operation = self.build_list_operation(**kwargs)
        response = await operation.execute()

        if 'error' in response:
            raise Exception(response['error'])

        models = response[self.model_class_name_plural]
        for model in models:
            match = find(
                lambda m: all([m[field] == model[field] for field in model_fields_to_check]),
                models,
            )
            self.assertEqual(match, model)

        return response

    async def base_test_execute_update(
        self,
        pk,
        payload: Dict,
        key_values_to_check: Dict,
        **kwargs,
    ):
        operation = self.build_update_operation(
            pk,
            {
                self.model_class_name: payload,
            },
            **kwargs,
        )
        response = await operation.execute()

        if 'error' in response:
            raise Exception(response['error'])

        model = response[self.model_class_name]
        for key, value in key_values_to_check.items():
            self.assertEqual(model[key], value)

        return response

    async def assertRaisesAsync(self, error_class, func):
        error_raised = False

        try:
            await func()
        except Exception as err:
            if err.__class__.__name__ == error_class.__name__:
                error_raised = True
            else:
                raise err

        self.assertTrue(error_raised)



from mage_ai.api.operations import constants
from mage_ai.api.presenters.BasePresenter import BasePresenter


class IntegrationDestinationPresenter(BasePresenter):
    default_attributes = [
        'name',
        'templates',
        'uuid',
    ]


IntegrationDestinationPresenter.register_format(
    constants.CREATE,
    [
        'error_message',
        'success',
    ],
)



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy


class SessionPolicy(BasePolicy):
    pass


SessionPolicy.allow_actions([
    constants.CREATE,
], scopes=[
    OauthScope.CLIENT_PUBLIC,
])

SessionPolicy.allow_actions([
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

SessionPolicy.allow_read([
    'expires',
    'token',
    'user',
], scopes=[
    OauthScope.CLIENT_PUBLIC,
], on_action=[
    constants.CREATE,
])

SessionPolicy.allow_read([
    'expires',
    'token',
    'user',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

SessionPolicy.allow_write([
    'email',
    'password',
    'username',
], scopes=[
    OauthScope.CLIENT_PUBLIC,
], on_action=[
    constants.CREATE,
])



from mage_ai.data_preparation.preferences import get_preferences
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.data_preparation.shared.secrets import get_secret_value
from mage_ai.data_preparation.sync import GitConfig
from mage_ai.orchestration.db.models.oauth import User
from mage_ai.shared.logger import VerboseFunctionExec
from typing import Any, List
from urllib.parse import urlparse
import asyncio
import base64
import os
import subprocess

DEFAULT_SSH_KEY_DIRECTORY = os.path.expanduser('~/.ssh')
REMOTE_NAME = 'mage-repo'


class Git:
    def __init__(self, git_config: GitConfig) -> None:
        import git
        self.remote_repo_link = git_config.remote_repo_link
        self.repo_path = git_config.repo_path or os.getcwd()
        os.makedirs(self.repo_path, exist_ok=True)
        self.git_config = git_config
        try:
            self.repo = git.Repo(self.repo_path)
        except git.exc.InvalidGitRepositoryError:
            self.repo = git.Repo.init(self.repo_path)
            # need to commit something to initialize the repository
            self.commit('Initial commit')

        self.__set_git_config()

        try:
            self.repo.create_remote(REMOTE_NAME, self.remote_repo_link)
        except git.exc.GitCommandError:
            # if the remote already exists
            pass

        self.origin = self.repo.remotes[REMOTE_NAME]
        if self.remote_repo_link not in self.origin.urls:
            self.origin.set_url(self.remote_repo_link)

    @classmethod
    def get_manager(self, user: User = None) -> 'Git':
        preferences = get_preferences(user=user)
        git_config = GitConfig.load(config=preferences.sync_config)
        return Git(git_config)

    @property
    def current_branch(self) -> Any:
        return self.repo.git.branch('--show-current')

    @property
    def branches(self) -> List:
        return [head.name for head in self.repo.heads]

    async def check_connection(self) -> None:
        proc = self.repo.git.ls_remote(self.origin.name, as_process=True)
        ct = 0
        while ct < 20:
            return_code = proc.poll()
            if return_code is not None:
                proc.kill()
                break
            ct += 1
            await asyncio.sleep(0.5)

        if return_code is not None and return_code != 0:
            _, err = proc.communicate()
            message = (
                err.decode('UTF-8') if err
                else 'Error connecting to remote, make sure your SSH key is set up properly.'
            )
            raise Exception(message)

        if return_code is None:
            proc.kill()
            raise TimeoutError(
                "Connecting to remote timed out, make sure your SSH key is set up properly"
                " and your repository host is added as a known host. More information here:"
                " https://docs.mage.ai/developing-in-the-cloud/setting-up-git#5-add-github-com-to-known-hosts")  # noqa: E501

    def _run_command(self, command: str) -> None:
        proc = subprocess.Popen(args=command, shell=True)
        proc.wait()

    def _remote_command(func) -> None:
        '''
        Decorator method for commands that need to connect to the remote repo. This decorator
        will configure and test SSH settings before executing the Git command.
        '''
        def wrapper(self, *args, **kwargs):
            if not os.path.exists(DEFAULT_SSH_KEY_DIRECTORY):
                os.mkdir(DEFAULT_SSH_KEY_DIRECTORY, 0o700)
            pubk_secret_name = self.git_config.ssh_public_key_secret_name
            if pubk_secret_name:
                public_key_file = os.path.join(
                    DEFAULT_SSH_KEY_DIRECTORY,
                    f'id_rsa_{pubk_secret_name}.pub'
                )
                if not os.path.exists(public_key_file):
                    try:
                        public_key = get_secret_value(
                            pubk_secret_name,
                            repo_name=get_repo_path(),
                        )
                        if public_key:
                            with open(public_key_file, 'w') as f:
                                f.write(base64.b64decode(public_key).decode('utf-8'))
                            os.chmod(public_key_file, 0o600)
                    except Exception:
                        pass
            pk_secret_name = self.git_config.ssh_private_key_secret_name
            private_key_file = os.path.join(DEFAULT_SSH_KEY_DIRECTORY, 'id_rsa')
            if pk_secret_name:
                private_key_file = os.path.join(
                    DEFAULT_SSH_KEY_DIRECTORY,
                    f'id_rsa_{pk_secret_name}'
                )
                if not os.path.exists(private_key_file):
                    try:
                        private_key = get_secret_value(
                            pk_secret_name,
                            repo_name=get_repo_path(),
                        )
                        if private_key:
                            with open(private_key_file, 'w') as f:
                                f.write(base64.b64decode(private_key).decode('utf-8'))
                            os.chmod(private_key_file, 0o600)
                    except Exception:
                        pass

            git_ssh_cmd = f'ssh -i {private_key_file}'
            with self.repo.git.custom_environment(GIT_SSH_COMMAND=git_ssh_cmd):
                try:
                    asyncio.run(self.check_connection())
                except TimeoutError as err:
                    url = f'ssh://{self.git_config.remote_repo_link}'
                    hostname = urlparse(url).hostname
                    if hostname:
                        cmd = f'ssh-keyscan -t rsa {hostname} >> ~/.ssh/known_hosts'  # noqa: E501
                        self._run_command(cmd)
                        asyncio.run(self.check_connection())
                    else:
                        raise err
                func(self, *args, **kwargs)

        return wrapper

    @_remote_command
    def reset(self, branch: str = None) -> None:
        self.origin.fetch()
        if branch is None:
            branch = self.current_branch
        self.repo.git.reset('--hard', f'{self.origin.name}/{branch}')
        self.__pip_install()

    @_remote_command
    def push(self) -> None:
        self.repo.git.push(
            '--set-upstream',
            self.origin.name,
            self.current_branch
        )

    @_remote_command
    def pull(self) -> None:
        self.origin.pull(self.current_branch)
        self.__pip_install()

    def status(self) -> str:
        return self.repo.git.status()

    def commit(self, message, files: List[str] = None) -> None:
        if self.repo.index.diff(None) or self.repo.untracked_files:
            if files:
                for file in files:
                    self.repo.git.add(file)
            else:
                self.repo.git.add('.')
            self.repo.index.commit(message)

    def change_branch(self, branch) -> None:
        if branch in self.repo.heads:
            current = self.repo.heads[branch]
        else:
            current = self.repo.create_head(branch)

        current.checkout()

    def __set_git_config(self) -> None:
        if self.git_config.username:
            self.repo.config_writer().set_value(
                'user', 'name', self.git_config.username).release()
        if self.git_config.email:
            self.repo.config_writer().set_value(
                'user', 'email', self.git_config.email).release()
        self.repo.config_writer('global').set_value(
            'safe', 'directory', self.repo_path).release()

    def __pip_install(self) -> None:
        requirements_file = os.path.join(
            self.repo.working_dir, 'requirements.txt')

        with VerboseFunctionExec(
            f'Running "pip3 install -r {requirements_file}"',
            verbose=True,
        ):
            try:
                if os.path.exists(requirements_file):
                    cmd = f'pip3 install -r {requirements_file}'
                    self._run_command(cmd)
                print(f'Installing {requirements_file} completed successfully.')
            except Exception as err:
                print(f'Skip installing {requirements_file} due to error: {err}')
                pass



from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig
from typing import Dict
import traceback

ECS_CONTAINER_METADATA_URI_VAR = 'ECS_CONTAINER_METADATA_URI_V4'


@dataclass
class K8sResourceConfig(BaseConfig):
    cpu: str
    memory: str


@dataclass
class K8sExecutorConfig(BaseConfig):
    resource_limits: K8sResourceConfig = None
    resource_requests: K8sResourceConfig = None

    @classmethod
    def load(self, config_path: str = None, config: Dict = None):
        executor_config = super().load(config_path=config_path, config=config)
        if executor_config.resource_limits is not None and \
                type(executor_config.resource_limits) is dict:
            try:
                executor_config.resource_limits = K8sResourceConfig.load(
                    config=executor_config.resource_limits,
                )
            except Exception:
                traceback.print_exc()
                executor_config.resource_limits = None
        if executor_config.resource_requests is not None and \
                type(executor_config.resource_requests) is dict:
            try:
                executor_config.resource_requests = K8sResourceConfig.load(
                    config=executor_config.resource_requests,
                )
            except Exception:
                traceback.print_exc()
                executor_config.resource_requests = None
        return executor_config



class ApiError(Exception):
    UNAUTHORIZED_ACCESS = {
        'code': 403,
        'message': 'Unauthorized access.',
        'type': 'unauthorized_access',
    }

    RESOURCE_ERROR = {
        'code': 500,
        'message': 'API resource error.',
        'type': 'api_resource_error',
    }
    RESOURCE_INVALID = {
        'code': 402,
        'message': 'Record is invalid.',
        'type': 'record_invalid',
    }
    RESOURCE_NOT_FOUND = {
        'code': 404,
        'message': 'Record not found.',
        'type': 'record_not_found',
    }

    INVALID_API_KEY = {
        'code': 403,
        'message': 'Invalid API key.',
        'type': 'invalid_api_key',
    }
    EXPIRED_OAUTH_TOKEN = {
        'code': 401,
        'message': 'Expired OAuth token.',
        'type': 'expired_oauth_token',
    }
    INVALID_OAUTH_TOKEN = {
        'code': 401,
        'message': 'Invalid OAuth token.',
        'type': 'invalid_oauth_token',
    }

    def __init__(self, opts={}):
        self.code = opts.get('code')
        self.errors = opts.get('errors')
        self.message = opts.get('message')
        self.type = opts.get('type')



from mage_ai.orchestration.queue.queue_factory import QueueFactory
from typing import Callable, Union
from enum import Enum


class JobType(str, Enum):
    BLOCK_RUN = 'block_run'
    PIPELINE_RUN = 'pipeline_run'


class JobManager:
    def __init__(self):
        self.queue = QueueFactory.get_queue()

    def add_job(
        self,
        job_type: JobType,
        uid: Union[str, int],
        target: Callable,
        *args,
        **kwargs
    ):
        job_id = self.__job_id(job_type, uid)

        self.queue.enqueue(job_id, target, *args, **kwargs)

    def clean_up_jobs(self):
        self.queue.clean_up_jobs()

    def has_block_run_job(self, block_run_id):
        job_id = self.__job_id(JobType.BLOCK_RUN, block_run_id)
        return self.queue.has_job(job_id)

    def has_pipeline_run_job(self, pipeline_run_id):
        job_id = self.__job_id(JobType.PIPELINE_RUN, pipeline_run_id)
        return self.queue.has_job(job_id)

    def kill_block_run_job(self, block_run_id):
        print(f'Kill block run id: {block_run_id}')
        job_id = self.__job_id(JobType.BLOCK_RUN, block_run_id)
        return self.queue.kill_job(job_id)

    def kill_pipeline_run_job(self, pipeline_run_id):
        print(f'Kill pipeline run id: {pipeline_run_id}')
        job_id = self.__job_id(JobType.PIPELINE_RUN, pipeline_run_id)
        return self.queue.kill_job(job_id)

    def __job_id(self, job_type: JobType, uid: Union[str, int]):
        return f'{job_type}_{uid}'


job_manager = JobManager()



from mage_ai.io.base import BaseSQLDatabase, ExportWritePolicy, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from mage_ai.io.export_utils import (
    clean_df_for_export,
    infer_dtypes,
)
from pandas import DataFrame
from typing import Dict, List, Union
import clickhouse_connect


class ClickHouse(BaseSQLDatabase):
    """
    Handles data transfer betwee a ClickHouse data warehouse and the Mage app.
    """
    def __init__(self, **kwargs) -> None:
        """
        Initializes settings for connecting to a ClickHouse warehouse.

        To authenticate (and authorize) access to a ClickHouse warehouse,
        credentials, i.e., username and password, must be provided.

        All keyword arguments will be passed to the ClickHouse client.
        """
        if kwargs.get('verbose') is not None:
            kwargs.pop('verbose')
        super().__init__(verbose=kwargs.get('verbose', True))
        with self.printer.print_msg('Connecting to ClickHouse'):
            self.client = clickhouse_connect.get_client(**kwargs)

    @classmethod
    def with_config(cls, config: BaseConfigLoader) -> 'ClickHouse':
        """
        Initializes ClickHouse client from configuration loader

        Args:
            config (BaseConfigLoader): Configuration loader object
        """
        if ConfigKey.CLICKHOUSE_HOST not in config:
            raise ValueError(
                'No valid configuration settings found for ClickHouse. '
                'You must specify host.'
            )
        return cls(
            database=config[ConfigKey.CLICKHOUSE_DATABASE],
            host=config[ConfigKey.CLICKHOUSE_HOST],
            interface=config[ConfigKey.CLICKHOUSE_INTERFACE],
            password=config[ConfigKey.CLICKHOUSE_PASSWORD],
            port=config[ConfigKey.CLICKHOUSE_PORT],
            username=config[ConfigKey.CLICKHOUSE_USERNAME],
        )

    def execute(self, command_string: str, **kwargs) -> None:
        """
        Sends command to the connected ClickHouse warehouse.

        Args:
            command_string (str): Command to execute on the ClickHouse warehouse.
            **kwargs: Additional arguments to pass to command, such as configurations
        """
        with self.printer.print_msg(f'Executing query \'{command_string}\''):
            command_string = self._clean_query(command_string)
            self.client.command(command_string, **kwargs)

    def execute_query(
        self,
        query: str,
        parameters: Dict = None,
        **kwargs,
    ) -> DataFrame:
        """
        Sends query to the connected ClickHouse warehouse.

        Args:
            query (str): Query to execute on the ClickHouse warehouse.
            **kwargs: Additional arguments to pass to query, such as query configurations
        """
        query = self._clean_query(query)
        with self.printer.print_msg(f'Executing query \'{query}\''):
            result = self.client.query_df(query, parameters=parameters)

        return result

    def execute_queries(
        self,
        queries: List[str],
        query_variables: List[Dict] = None,
        fetch_query_at_indexes: List[bool] = None,
        **kwargs,
    ) -> List:
        results = []

        for idx, query in enumerate(queries):
            parameters = query_variables[idx] \
                        if query_variables and idx < len(query_variables) \
                        else {}
            query = self._clean_query(query)

            if fetch_query_at_indexes and idx < len(fetch_query_at_indexes) and \
                    fetch_query_at_indexes[idx]:
                result = self.client.query_df(query, parameters=parameters)
            else:
                result = self.client.command(query, parameters=parameters)

            results.append(result)

        return results

    def load(
        self,
        query_string: str,
        limit: int = QUERY_ROW_LIMIT,
        display_query: Union[str, None] = None,
        verbose: bool = True,
        **kwargs,
    ) -> DataFrame:
        """
        Loads data from ClickHouse into a Pandas data frame based on the query given.
        This will fail if the query returns no data from the database. When a select query
        is provided, this function will load at maximum 10,000,000 rows of data. To operate on more
        data, consider performing data transformations in warehouse.

        Args:
            query_string (str): Query to fetch a table or subset of a table.
            limit (int, Optional): The number of rows to limit the loaded dataframe to. Defaults to
                                    10,000,000.
            **kwargs: Additional arguments to pass to query, such as query configurations

        Returns:
            DataFrame: Data frame associated with the given query.
        """

        print_message = 'Loading data'
        if verbose:
            print_message += ' with query'

            if display_query:
                for line in display_query.split('\n'):
                    print_message += f'\n{line}'
            else:
                print_message += f'\n{query_string}'

        query_string = self._clean_query(query_string)

        with self.printer.print_msg(print_message):
            return self.client.query_df(
                self._enforce_limit(query_string, limit), **kwargs
            )

    def export(
        self,
        df: DataFrame,
        table_name: str,
        database: str = 'default',
        if_exists: str = 'append',
        index: bool = False,
        query_string: Union[str, None] = None,
        create_table_statement: Union[str, None] = None,
        verbose: bool = True,
        **kwargs,
    ) -> None:
        """
        Exports a Pandas data frame to a ClickHouse warehouse based on the table name.
        If table doesn't exist, the table is automatically created.

        Args:
            df (DataFrame): Data frame to export
            table_name (str): Name of the table to export data to (excluding database).
            If this table exists, the table schema must match the data frame schema.
            If this table doesn't exist, query_string must be specified to create the new table.
            database (str): Name of the database in which the table is located.
            if_exists (str, optional): Specifies export policy if table exists. Either
                - `'fail'`: throw an error.
                - `'replace'`: drops existing table and creates new table of same name.
                - `'append'`: appends data frame to existing table. In this case the schema must
                                match the original table.
            Defaults to `'append'`.
            **kwargs: Additional arguments to pass to writer
        """

        if type(df) is dict:
            df = DataFrame([df])
        elif type(df) is list:
            df = DataFrame(df)

        if not query_string:
            if index:
                df = df.reset_index()

            dtypes = infer_dtypes(df)
            df = clean_df_for_export(df, self.clean, dtypes)

        def __process(database: Union[str, None]):

            df_existing = self.client.query_df(f"""
EXISTS TABLE {database}.{table_name}
""")

            table_exists = not df_existing.empty and df_existing.iloc[0, 0] == 1
            should_create_table = not table_exists

            if table_exists:
                if ExportWritePolicy.FAIL == if_exists:
                    raise ValueError(
                        f'Table \'{table_name}\' already'
                        ' exists in database {database}.',
                    )
                elif ExportWritePolicy.REPLACE == if_exists:
                    self.client.command(
                        f'DROP TABLE IF EXISTS {database}.{table_name}')
                    should_create_table = True

            if query_string:
                self.client.command(f'USE {database}')

                if should_create_table:
                    self.client.command(f"""
CREATE TABLE IF NOT EXISTS {database}.{table_name} ENGINE = Memory AS
{query_string}
""")
                else:
                    self.client.command(f"""
INSERT INTO {database}.{table_name}
{query_string}
""")
            else:
                if should_create_table:
                    self.client.command(create_table_statement)

                self.client.insert_df(f'{database}.{table_name}', df)

        if verbose:
            with self.printer.print_msg(
                    f'Exporting data to table \'{database}.{table_name}\''):
                __process(database=database)
        else:
            __process(database=database)



from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.models.pipelines.integration_pipeline import IntegrationPipeline
from mage_ai.orchestration.db import safe_db_query


class IntegrationSourceStreamResource(GenericResource):
    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        return self(IntegrationPipeline.get(pk), user, **kwargs)



from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.base import DataSource
from mage_ai.orchestration.db import safe_db_query
import aiofiles
import yaml


DATA_PROVIDERS = [
    DataSource.BIGQUERY,
    DataSource.CLICKHOUSE,
    DataSource.MSSQL,
    DataSource.MYSQL,
    DataSource.POSTGRES,
    DataSource.REDSHIFT,
    DataSource.SNOWFLAKE,
    DataSource.TRINO,
]
DATA_PROVIDERS_NAME = {
    DataSource.BIGQUERY: 'BigQuery',
    DataSource.CLICKHOUSE: 'ClickHouse',
    DataSource.MSSQL: 'Microsoft SQL Server',
    DataSource.MYSQL: 'MySQL',
    DataSource.POSTGRES: 'PostgreSQL',
    DataSource.REDSHIFT: 'Redshift',
    DataSource.SNOWFLAKE: 'Snowflake',
    DataSource.TRINO: 'Trino',
}


class DataProviderResource(GenericResource):
    @classmethod
    @safe_db_query
    async def collection(self, query, meta, user, **kwargs):
        async with aiofiles.open(f'{get_repo_path()}/io_config.yaml', mode='r') as file:
            try:
                profiles = list(yaml.safe_load(await file.read()).keys())
            except yaml.YAMLError as exc:
                print(exc)

        collection = [dict(
            id=DATA_PROVIDERS_NAME[ds.value],
            profiles=[p for p in profiles if p != 'version'],
            value=ds.value,
        ) for ds in DATA_PROVIDERS]

        return self.build_result_set(
            collection,
            user,
            **kwargs,
        )



EMAIL_NOTIFICATION_CONFIG = dict(
    alert_on=[
        'trigger_failure',
        'trigger_success',
    ],
    email_config=dict(
        smtp_host='test_host',
        smtp_mail_from='test_from@abc.com',
        smtp_user='test_user',
        smtp_password='test_password',
        to_emails=['test_to@xyz.com'],
    )
)

SLACK_NOTIFICATION_CONFIG = dict(
    alert_on=[
        'trigger_failure',
        'trigger_success',
    ],
    slack_config=dict(
        webhook_url='test_webhook_url',
    )
)

GOOGLE_CHAT_NOTIFICATION_CONFIG = dict(
    alert_on=[
        'trigger_failure',
        'trigger_success',
    ],
    google_chat_config=dict(
        webhook_url='test_webhook_url',
    )
)

TEAMS_NOTIFICATION_CONFIG = dict(
    alert_on=[
        'trigger_failure',
        'trigger_success',
    ],
    teams_config=dict(
        webhook_url='test_webhook_url',
    )
)


TEAMS_NOTIFICATION_CONFIG_NO_ALERT_ON = dict(
    alert_on=[],
    teams_config=dict(
        webhook_url='test_webhook_url',
    )
)



CHART_TYPE_BAR_HORIZONTAL = 'bar_horizontal'
CHART_TYPE_LINE_CHART = 'line_chart'
CHART_TYPE_HISTOGRAM = 'histogram'

LABEL_TYPE_RANGE = 'range'

DATA_KEY_CHARTS = 'charts'
DATA_KEY_CORRELATION = 'correlations'
DATA_KEY_OVERVIEW = 'overview'
DATA_KEY_SCATTER_PLOT = 'scatter_plot'
DATA_KEY_SCATTER_PLOT_LABELS = 'scatter_plot_labels'
DATA_KEY_TIME_SERIES = 'time_series'
DATA_KEYS = [
    DATA_KEY_CHARTS,
    DATA_KEY_CORRELATION,
    DATA_KEY_OVERVIEW,
    DATA_KEY_SCATTER_PLOT,
    DATA_KEY_TIME_SERIES,
]



{% extends "transformers/transformer_actions/action.jinja" %}
{% block imports %}
{{ super() -}}
{% endblock %}
{% block action %}
    """
    Execute Transformer Action: ActionType.STANDARDIZE

    Docs: https://docs.mage.ai/guides/transformer-blocks#standardize-data
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.STANDARDIZE,
        arguments=[],  # Specify columns to normalize
        axis=Axis.COLUMN, # Specify normalization strategy
    )
{% endblock %}



from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.orchestration.db import safe_db_query


class EventRuleResource(GenericResource):
    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        rules = []

        if 'aws' == pk:
            from mage_ai.services.aws.events.events import get_all_event_rules
            rules = get_all_event_rules()

        return self(dict(rules=rules), user, **kwargs)



from mage_ai.api.operations import constants
from mage_ai.api.presenters.BasePresenter import BasePresenter
from mage_ai.data_preparation.models.constants import DATAFRAME_SAMPLE_COUNT_PREVIEW


class PipelinePresenter(BasePresenter):
    default_attributes = [
        'blocks',
        'data_integration',
        'description',
        'executor_count',
        'name',
        'type',
        'updated_at',
        'uuid',
        'variables',
        'widgets',
    ]

    async def present(self, **kwargs):
        query = kwargs.get('query', {})
        display_format = kwargs['format']

        include_extensions = query.get('includes_extensions', [True])
        if include_extensions:
            include_extensions = include_extensions[0]

        if constants.DETAIL == display_format:
            include_content = query.get('includes_content', [True])
            if include_content:
                include_content = include_content[0]

            include_outputs = query.get('includes_outputs', [True])
            if include_outputs:
                include_outputs = include_outputs[0]

            include_block_metadata = query.get('includes_block_metadata', [True])
            if include_block_metadata:
                include_block_metadata = include_block_metadata[0]

            return await self.model.to_dict_async(
                include_block_metadata=include_block_metadata,
                inclide_block_tags=True,
                include_content=include_content,
                include_extensions=include_extensions,
                include_outputs=include_outputs,
                sample_count=DATAFRAME_SAMPLE_COUNT_PREVIEW,
            )
        elif constants.UPDATE == display_format:
            data = self.model.to_dict(include_extensions=include_extensions)
        else:
            data = self.model.to_dict()

        include_schedules = query.get('include_schedules', [False])
        if include_schedules:
            include_schedules = include_schedules[0]

        if include_schedules:
            data['schedules'] = self.model.schedules

        return data


PipelinePresenter.register_format(
    constants.LIST,
    PipelinePresenter.default_attributes + [
        'schedules',
    ],
)



from datetime import datetime
from faker import Faker
from mage_ai.authentication.passwords import create_bcrypt_hash, generate_salt
from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.orchestration.db.models.oauth import User
from mage_ai.orchestration.db.models.schedules import (
    PipelineRun,
    PipelineSchedule,
)
from mage_ai.shared.hash import merge_dict
from typing import Dict, Union

faker = Faker()


def create_pipeline(name: str, repo_path: str):
    pipeline = Pipeline.create(
        name,
        repo_path=repo_path,
    )
    return pipeline


def create_pipeline_with_blocks(name: str, repo_path: str):
    pipeline = Pipeline.create(
        name,
        repo_path=repo_path,
    )
    block1 = Block.create('block1', 'data_loader', repo_path, language='python')
    block2 = Block.create('block2', 'transformer', repo_path, language='python')
    block3 = Block.create('block3', 'transformer', repo_path, language='python')
    block4 = Block.create('block4', 'data_exporter', repo_path, language='python')
    pipeline.add_block(block1)
    pipeline.add_block(block2, upstream_block_uuids=['block1'])
    pipeline.add_block(block3, upstream_block_uuids=['block1'])
    pipeline.add_block(block4, upstream_block_uuids=['block2', 'block3'])
    return pipeline


def create_pipeline_with_dynamic_blocks(name: str, repo_path: str):
    pipeline = Pipeline.create(
        name,
        repo_path=repo_path,
    )
    block1 = Block.create(
        'block1',
        'data_loader',
        repo_path,
        language='python',
        configuration=dict(dynamic=True),
    )
    block2 = Block.create(
        'block2', 'transformer', repo_path, language='python')
    block3 = Block.create(
        'block3', 'data_exporter', repo_path, language='python')
    pipeline.add_block(block1)
    pipeline.add_block(block2, upstream_block_uuids=['block1'])
    pipeline.add_block(block3, upstream_block_uuids=['block2'])
    return pipeline


def create_pipeline_run(pipeline_uuid: str):
    pipeline_run = PipelineRun.create(pipeline_uuid='test_pipeline')
    return pipeline_run


def create_pipeline_run_with_schedule(
    pipeline_uuid: str,
    execution_date: datetime = None,
    pipeline_schedule_id: int = None,
    pipeline_schedule_settings: Dict = dict(),
):
    if pipeline_schedule_id is None:
        pipeline_schedule = PipelineSchedule.create(
            pipeline_uuid=pipeline_uuid,
            settings=pipeline_schedule_settings,
        )
        pipeline_schedule_id = pipeline_schedule.id
    pipeline_run = PipelineRun.create(
        execution_date=execution_date,
        pipeline_uuid=pipeline_uuid,
        pipeline_schedule_id=pipeline_schedule_id,
    )
    return pipeline_run


def create_user(
    as_dict: bool = False,
    save: bool = True,
    password: str = None,
    **kwargs,
) -> Union[Dict, User]:
    password = password or faker.password()
    password_salt = generate_salt()
    password_hash = create_bcrypt_hash(password, password_salt)
    payload = merge_dict(dict(
        email=faker.email(),
        username=faker.name(),
    ), kwargs)

    if as_dict:
        payload.update(password=password)
        return payload

    user = User(
        password_hash=password_hash,
        password_salt=password_salt,
        **payload,
    )
    if save:
        user.save()

    return user



from mage_ai.api.presenters.BasePresenter import BasePresenter


class SyncPresenter(BasePresenter):
    default_attributes = [
        'branch',
        'remote_repo_link',
        'username',
        'email',
        'ssh_private_key',
        'ssh_public_key',
        'repo_path',
        'sync_on_pipeline_run',
        'type',
    ]



from collections import UserList
from collections.abc import Iterable
from datetime import datetime
from functools import reduce
from mage_ai.api.operations.constants import READ
from mage_ai.api.resources.BaseResource import BaseResource
from mage_ai.orchestration.db.models.base import BaseModel
from mage_ai.shared.hash import merge_dict
import importlib
import inspect


class BasePresenter():
    all_attributes_attr = {}
    all_formats_attr = {}
    default_attributes = []

    def __init__(self, resource, current_user, **kwargs):
        self.current_user = current_user
        self.options = kwargs
        self.resource = resource

    @classmethod
    def all_attributes(self):
        if not self.all_attributes_attr.get(self.__name__):
            self.all_attributes_attr[self.__name__] = {}
        return self.all_attributes_attr[self.__name__]

    @classmethod
    def all_formats(self):
        if not self.all_formats_attr.get(self.__name__):
            self.all_formats_attr[self.__name__] = {
                'default': self.default_attributes,
            }
        return self.all_formats_attr[self.__name__]

    @classmethod
    def formats(self, format_arg):
        if format_arg and self.all_formats().get(format_arg, None) is not None:
            return self.all_formats()[format_arg]
        else:
            return self.all_formats()['default']

    @classmethod
    def register_attributes(self, keys, klass_symbol_or_lambda):
        for key in keys:
            self.all_attributes()[key] = klass_symbol_or_lambda

    @classmethod
    def register_format(self, format_arg, keys):
        self.all_formats()[format_arg] = keys

    @classmethod
    def register_formats(self, formats, keys):
        arr = formats if isinstance(formats, list) else [formats]
        for format_arg in arr:
            self.register_format(format_arg, keys)

    @classmethod
    async def present_resource(self, resource, user, **kwargs):
        async def present_lambda(r):
            if r and inspect.isawaitable(r):
                r = await r

            results = r.__class__.presenter_class()(
                r,
                user,
                **kwargs,
            ).present(
                **kwargs,
            )

            if results and inspect.isawaitable(results):
                results = await results

            return results

        if isinstance(resource, Iterable):
            return [await present_lambda(r) for r in resource]
        else:
            return await present_lambda(resource)

    @classmethod
    def present_model(self, model, resource_class, user, **kwargs):
        if model:
            return self.present_resource(
                resource_class(model, user, **kwargs),
                user,
                **kwargs,
            )

    @classmethod
    def present_models(self, models, resource_class, user, **kwargs):
        return self.present_resource(
            resource_class.build_result_set(models, user, **kwargs),
            user,
            **kwargs,
        )

    async def present(self, **kwargs):
        def _build(obj, key):
            value = getattr(self, key)
            if callable(value):
                value = value(**kwargs)
            self.__validate_attribute_type(key, value)
            if issubclass(
                    value.__class__,
                    list) or issubclass(
                    value.__class__,
                    UserList):
                obj[key] = [
                    self.__transform_value(
                        key, v, **kwargs) for v in value]
            else:
                obj[key] = self.__transform_value(key, value, **kwargs)
            return obj

        format_to_present = kwargs.get('format', None)
        if format_to_present and self.options.get('from_resource'):
            from_resource_name = self.options['from_resource'].resource_name_singular(
            )
            format_to_present = f'{from_resource_name}/{format_to_present}'

        return reduce(_build, self.__class__.formats(format_to_present), {})

    def __transform_value(self, key, value, **kwargs):
        klass_symbol_or_lambda = self.__class__.all_attributes().get(key, None)

        if issubclass(value.__class__, BaseModel):
            resource_class_name = f'{value.__class__.__name__}Resource'
            resource_class = getattr(importlib.import_module(
                f'mage_ai.api.resources.{resource_class_name}'), resource_class_name, )
            value = resource_class(value, self.current_user, **kwargs)

        if isinstance(value, datetime):
            return str(value)
        elif klass_symbol_or_lambda is float:
            return float(value)
        elif klass_symbol_or_lambda is int:
            return int(value)
        elif issubclass(value.__class__, BaseResource):
            opts = self.options.copy()
            opts['from_resource'] = self.resource
            data = value.presenter_class().present_resource(
                value,
                self.current_user,
                **merge_dict(kwargs, opts),
            )

            if not kwargs.get('ignore_permissions'):
                policy = value.policy_class()(value, self.current_user, **opts)
                policy.authorize_attributes(
                    READ,
                    data.keys(),
                    **opts,
                )

            return data
        else:
            return value

    def __validate_attribute_class(self, klass_symbol, value):
        pass

    def __validate_attribute_type(self, key, value):
        pass

    def __getattr__(self, name):
        def _missing(*args, **kwargs):
            val = getattr(self.resource, name)
            if callable(val):
                return val(*args, **kwargs)
            else:
                return val
        return _missing()



from collections import OrderedDict
from mage_ai.data_cleaner.estimators.base import BaseEstimator
from mage_ai.shared.conversions import fd_to_df, np_to_fd
from mage_ai.shared.multi import execute_parallel
from sklearn.preprocessing import LabelEncoder
import numpy as np


class CustomLabelEncoder(BaseEstimator):
    def __init__(self):
        self.encoder = LabelEncoder()
        self.fitted = False
        self.label_values = []
        self.unknown_class = None

    def inverse_transform(self, y):
        return self.encoder.inverse_transform(y)

    def fit(self, X, y=None):
        self.encoder.fit(X)
        self.fitted = True
        return self

    def transform(self, X, **kwargs):
        self.label_values = sorted(set(X))
        class_mappings = dict(
            zip(
                [str(c) for c in self.encoder.classes_],
                self.encoder.transform(self.encoder.classes_),
            ),
        )
        missing_value = len(self.encoder.classes_)
        unknown_found = False

        def _build(x):
            v = class_mappings.get(str(x))
            if v is None:
                return missing_value
            return v
        if unknown_found:
            # TODO(christhetree): why are these multiplied by 2?
            if np.issubdtype(X.dtype, np.floating):
                self.unknown_class = float(self.label_values[-1] * 2)
            elif np.issubdtype(X.dtype, np.integer):
                self.unknown_class = int(self.label_values[-1] * 2)
            else:
                self.unknown_class = 'unknown_class_'
        y = np.array([_build(x) for x in X])
        return y

    def label_classes(self):
        if self.unknown_class:
            return self.encoder.classes_ + [self.unknown_class]
        return self.encoder.classes_


class MultipleColumnLabelEncoder(BaseEstimator):
    def __init__(self, input_type=None):
        self.input_type = input_type
        self.encoders = {}

    def fit(self, X, y=None):
        execute_parallel(
            [(self.fit_column, [X[column], column]) for column in X.columns],
        )
        return self

    def fit_column(self, X, column):
        self.encoders[column] = CustomLabelEncoder()
        if self.input_type:
            self.encoders[column].fit(X.apply(self.input_type))
        else:
            self.encoders[column].fit(X)

    def transform(self, X, parallel=False):
        if parallel:
            output_dict = OrderedDict()
            output_dicts = execute_parallel(
                [(self.transform_column, [X[column], column]) for column in X.columns],
            )

            for od in output_dicts:
                output_dict.update(od)
            return fd_to_df(output_dict)

        X_output = X.copy(deep=True)
        for col in X_output.columns:
            if self.input_type:
                X_output[col] = self.encoders[col].transform(
                    X_output[col].apply(self.input_type),
                )
            else:
                X_output[col] = self.encoders[col].transform(X_output[col])
        return X_output

    def transform_column(self, X, column):
        if self.input_type:
            nd_arr = self.encoders[column].transform(X.apply(self.input_type))
        else:
            nd_arr = self.encoders[column].transform(X)
        return np_to_fd(nd_arr, feature_names=[column])



from abc import ABC, abstractmethod
from typing import Tuple
from ldap3 import Server, Connection
from ldap3.core.exceptions import LDAPException
from mage_ai.settings import (
    LDAP_SERVER,
    LDAP_BIND_DN,
    LDAP_BIND_PASSWORD,
    LDAP_BASE_DN,
    LDAP_AUTHENTICATION_FILTER,
    LDAP_AUTHORIZATION_FILTER,
)


class LDAPAuthenticator(ABC):
    @abstractmethod
    def authenticate(self, username: str, password: str) -> Tuple[bool, str]:
        pass

    @abstractmethod
    def authorize(self, username: str) -> bool:
        pass

    def verify(self, username: str, password: str) -> bool:
        authenticated, user_dn = self.authenticate(username, password)
        if authenticated:
            return self.authorize(user_dn)
        return False


class LDAPConnection(LDAPAuthenticator):
    def __init__(
        self,
        server_url,
        bind_dn,
        bind_password,
        base_dn,
        authentication_filter,
        authorization_filter,
    ):
        self.server_url = server_url
        self.bind_dn = bind_dn
        self.bind_password = bind_password
        self.base_dn = base_dn
        self.authentication_filter = authentication_filter
        self.authorization_filter = authorization_filter
        self.conn = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.conn.unbind()

    def bind(self):
        self.conn = Connection(
            Server(self.server_url), self.bind_dn, self.bind_password, auto_bind=True
        )

    def authenticate(self, username: str, password: str) -> Tuple[bool, str]:
        try:
            if not self.conn:
                self.bind()
            self.conn.search(
                self.base_dn, self.authentication_filter.format(username=username)
            )
            if not self.conn.entries:
                return False, ""
            user_dn = self.conn.entries[0].entry_dn
            if self.conn.rebind(user=user_dn, password=password):
                return True, user_dn
            return False, user_dn
        except LDAPException:
            return False, ""

    def authorize(self, user_dn: str) -> bool:
        try:
            if not self.conn:
                self.bind()
            self.conn.entries.clear()
            self.conn.search(self.base_dn, self.authorization_filter.format(user_dn=user_dn))
            if self.conn.entries:
                if len(self.conn.entries) > 0:
                    return True
            return False
        except LDAPException:
            return False


def new_ldap_connection() -> LDAPConnection:
    return LDAPConnection(
        LDAP_SERVER,
        LDAP_BIND_DN,
        LDAP_BIND_PASSWORD,
        LDAP_BASE_DN,
        LDAP_AUTHENTICATION_FILTER,
        LDAP_AUTHORIZATION_FILTER,
    )


def verify(username, password: str) -> bool:
    return LDAPConnection(
        LDAP_SERVER,
        LDAP_BIND_DN,
        LDAP_BIND_PASSWORD,
        LDAP_BASE_DN,
        LDAP_AUTHENTICATION_FILTER,
        LDAP_AUTHORIZATION_FILTER,
    ).verify(username, password)



from datetime import datetime
from mage_ai.services import datadog as dd
from mage_ai.tests.base_test import TestCase
from unittest.mock import Mock, patch

TEST_METRIC = 'mage.test'
TAGS = dict(tag1='tag')


class DatadogTests(TestCase):
    @patch('datadog.api.Event.create')
    def test_create_event(self, mock_event):
        event_name = 'event name'
        event_text = 'event text'
        dd.create_event(event_name, event_text, tags=TAGS)
        mock_event.assert_called_with(title=event_name, text=event_text, tags=TAGS)

    @patch('datadog.api.Metric.send')
    @patch('mage_ai.services.datadog.datetime')
    def test_gauge(self, mock_dt, mock_metric):
        dt = datetime(2023, 1, 1)
        mock_dt.utcnow = Mock(return_value=dt)
        dd.gauge(TEST_METRIC, 5, tags=TAGS)
        mock_metric.assert_called_with(
            host='mage',
            metric=TEST_METRIC,
            points=[(dt.timestamp(), 5)],
            tags=TAGS,
            type='gauge'
        )

    @patch('datadog.api.Metric.send')
    def test_increment(self, mock_metric):
        dd.increment(TEST_METRIC, tags=TAGS)
        mock_metric.assert_called_with(metrics=[{
            'host': 'mage',
            'metric': TEST_METRIC,
            'points': 1,
            'tags': TAGS,
            'type': 'count'
        }])

    @patch('datadog.api.Metric.send')
    @patch('mage_ai.services.datadog.datetime')
    def test_histogram(self, mock_dt, mock_metric):
        dt = datetime(2023, 1, 1)
        mock_dt.utcnow = Mock(return_value=dt)
        dd.histogram(TEST_METRIC, 3, tags=TAGS)
        mock_metric.assert_called_with(
            host='mage',
            metric=TEST_METRIC,
            points=[(dt.timestamp(), 3)],
            tags=TAGS,
            type='histogram'
        )

    @patch('datadog.api.Metric.send')
    @patch('mage_ai.services.datadog.datetime')
    def test_timing(self, mock_dt, mock_metric):
        dt = datetime(2023, 1, 1)
        mock_dt.utcnow = Mock(return_value=dt)
        dd.timing(TEST_METRIC, 100, tags=TAGS)
        mock_metric.assert_called_with(
            host='mage',
            metric=TEST_METRIC,
            points=[(dt.timestamp(), 100)],
            tags=TAGS,
            type='timer'
        )



from mage_ai.data_cleaner.transformer_actions.variable_replacer import (
    interpolate,
    replace_true_false,
)
from mage_ai.tests.base_test import TestCase
from mage_ai.tests.data_cleaner.transformer_actions.shared import TEST_ACTION


class VariableReplacerTests(TestCase):
    def test_interpolate(self):
        text = TEST_ACTION['action_code']
        key1 = '1_1'
        variable_data1 = TEST_ACTION['action_variables'][key1]
        key2 = '1_2'
        variable_data2 = TEST_ACTION['action_variables'][key2]
        key3 = '1'
        variable_data3 = TEST_ACTION['action_variables'][key3]

        self.assertEqual(
            interpolate(
                interpolate(interpolate(text, key1, variable_data1), key2, variable_data2),
                key3,
                variable_data3,
            ),
            'omni.deposited == True and (omni.fund == "The Quant" or omni.fund == "Yield")',
        )

    def test_replace_true_false(self):
        action_code = 'a == false and b == true or (a == true and b == false) and ' \
            'a == False and b == True or a == "true" and b == "false" or ' \
            "a == 'false' and b == 'true' or a == 'True' and b == 'False'"
        result = 'a == False and b == True or (a == True and b == False) and ' \
            'a == False and b == True or a == "true" and b == "false" or ' \
            "a == 'false' and b == 'true' or a == 'True' and b == 'False'"
        self.assertEqual(replace_true_false(action_code), result)
